import zelas2.Multispectral
import zelas2.voxel
import zelas2.TheHeartOfTheMilitaryGod
import zelas2.RedundancyElimination
import zelas2.Ellipse
import zelas2.ransac
import zelas2.shield
import zelas2.dxf_dwg
import zelas2.graphCut
import zelas2.mmdet3d_base
import zelas2.shield_plus



#coding=utf-8
import numpy as np  # numpy库
import math  # 添加数学库
import matplotlib.pyplot as plt  # 显示库
import open3d as o3d  # open3d库
from mpl_toolkits.mplot3d import Axes3D  # 空间三维画图
import scipy.spatial as spt  # 求凸壳用的库
from shapely.geometry import Polygon  # 求凸壳面积用的库
from sklearn.mixture import GaussianMixture  # GMM库
import time as s  # 时间库
import multiprocessing as mp  # 添加多线程库
from astropy.modeling import models, fitting  # 添加单高斯拟合函数

# 求面积（平面面积以及侧面积）
def get_area(xy):
    hull = spt.ConvexHull(points=xy, incremental=False)  # 求凸壳
    ID = hull.vertices  # 返回凸壳的边缘点号
    polygon = xy[ID, :]  # 求凸壳数组
    area = Polygon(polygon).area  # 求多边形面积
    return area

# 求每个点所占的平均面积
def mean2Ddensity(num, Area):
    density = math.sqrt(Area/num)
    print('平面平均点间距为:', density)
    return density

# https://www.cnblogs.com/wt714/p/12545129.html
# 对强度值数集进行归一化
def normalization(intensity, threshold=1):
    max_i = max(intensity)
    min_i = min(intensity)
    new_intensity = (intensity-min_i)/(max_i - min_i)*threshold
    return new_intensity

def normalization2(arr_i,max_i=1,min_i=0):
    '特征值标准化映射算法'
    # 获取数组的最小值和最大值
    min_val = np.min(arr_i)
    max_val = np.max(arr_i)
    # 使用线性插值进行映射
    mapped_array = ((arr_i - min_val) / (max_val - min_val)) * (max_i - min_i) + min_i
    return mapped_array

#求两个航带之间的最邻近点,并返回求最邻近加权后的同点强度值
def kntree_point_num(knxyz, xyz,kn_i,i,max_n=5):
    c_o3d = o3d.geometry.PointCloud()  # 建立open3d类，准备进行最邻近
    o3d_xyz = np.column_stack((knxyz[:, 0], knxyz[:, 1], knxyz[:, 2])).astype(np.float32)
    c_o3d.points = o3d.utility.Vector3dVector(o3d_xyz)  # 将点云和容器进行连接
    pcd_tree = o3d.geometry.KDTreeFlann(c_o3d)  # 对第二行带建立kntree
    i_new = []  # 新强度值存储容器
    for i in range(len(xyz)):  # 对整个点云进行遍历
        [_, idx_kn, dis_kn] = pcd_tree.search_knn_vector_3d(xyz[i, :], max_n)
        i_kn = kn_i[idx_kn]  # 存储最邻近的几个点的强度值
        dis_t = 1/np.array(dis_kn)  # 反距离加权法计算每个点的权值
        i_new_singel = np.dot(i_kn, dis_t)/sum(dis_t)  # 数乘求出最后的新的强度值
        i_new.append(i_new_singel)  # 将新的强度值集合到容器中
    return np.array(i_new)  # 返回数组类型的点云强度值

# 建立频率域直方图
def histogram(i1, i2, i3,bin=300,d=True):
        plt.rcParams['font.sans-serif'] = ['SimHei']  # 解决中文无法显示的问题
        plt.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号
        plt.subplot(311)  # 3行1列第一幅图
        plt.hist(sorted(i1), bins=bin,density=d)  # bins表示分为5条直方，可以根据需求修改
        plt.xlabel('intensity')
        plt.ylabel('frequency')
        #plt.xlim( 0, 50 )
        plt.subplot(312)  # 3行1列第二幅图
        plt.hist(sorted(i2), bins=bin, density=d)  # density为True表示频率，否则是频数，可根据需求修改
        plt.xlabel('intensity')
        plt.ylabel('frequency')
        #plt.xlim(0, 50)
        plt.subplot(313)  # 3行1列第三幅图
        plt.hist(sorted(i3), bins=bin, density=d)  # density为True表示频率，否则是频数，可根据需求修改
        plt.xlabel('intensity')
        plt.ylabel('frequency')
        #plt.xlim(0, 50)
        plt.show()

# 点云噪声剔除函数
def NoiseElimination(xyzi, max_i):
    id = []  # 一个删除点云下标的容器
    for i in range(len(xyzi)):  # 对点云进行遍历
        if xyzi[i, 3] > max_i: # 判断是否超过阈值
            id.append(i)  # 记录点云下标
    print('剔除强度粗差',len(id),'个点')
    xyzi = np.delete(xyzi, id, axis=0)  # 删除符合条件的点云
    return xyzi  # 返回处理后的点云

# 以强度值为基础，映射到RBG，进行三维散点图的显示
def i32rgb(xyz,i1,i2,i3,max_i=255):  # 输入点云位置，以及映射的三个点云强度值
    fig = plt.figure() # 新建一个窗口
    ax = Axes3D(fig) # 明确为三维散点图
    if max(i1)+max(i2)+max(i3) >3: # 如果没有进行归一化,那么进行归一化
        i1 /= max_i
        i2 /= max_i
        i3 /= max_i
    rgb = np.c_[i1, i2, i3]  # 将三强度值附给rgb三个通道
    ax.scatter(xyz[:,0], xyz[:,1], xyz[:,2], c=rgb)  # 显示点云
    plt.show() # 显示窗口

# 输入点云三维坐标，返回点云的体素
def max_min(xyzi_3,density_xy,S=0):
    x = xyzi_3[:, 0]
    y = xyzi_3[:, 1]
    z = xyzi_3[:, 2]  # x y z 分离
    id_x_max = np.argmax(x)
    id_x_min = np.argmin(x)
    id_y_max = np.argmax(y)
    id_y_min = np.argmin(y)
    id_z_max = np.argmax(z)
    id_z_min = np.argmin(z)  # 找到极值点云所在位置
    # 建立一个存储极值点的矩阵
    cuboid = np.empty((6,3))
    cuboid[0, :] = xyzi_3[id_x_max, :3]
    cuboid[1, :] = xyzi_3[id_x_min, :3] # 最小x
    cuboid[2, :] = xyzi_3[id_y_max, :3]
    cuboid[3, :] = xyzi_3[id_y_min, :3] # 最小y
    cuboid[4, :] = xyzi_3[id_z_max, :3]
    cuboid[5, :] = xyzi_3[id_z_min, :3] # 最小z
    l_xyz = np.empty((3))
    l_xyz[0] = cuboid[0, 0]-cuboid[1, 0]
    l_xyz[1] = cuboid[2, 1] - cuboid[3, 1]
    l_xyz[2] = cuboid[4, 2] - cuboid[5, 2]  # 求边长
    num_dis = np.ceil(l_xyz/density_xy)

    # 新建一个体素三维数组
    voxel = np.zeros((round(num_dis[0]),round(num_dis[1]),round(num_dis[2]),4))  # 长*宽*高*强度值*3*点云数量
    voxel[:,:,:,3] = 0 # 计算包含点云数量的容器初始化
    min_p = [min(x), min(y), min(z)]  # 存储最小值
    # 给每个点云进行标记
    LWH = np.empty((len(xyzi_3),3)) # 新建一个容器用来存储每个点云属于哪个体元
    for i in range(len(xyzi_3)):
        xi = xyzi_3[i,0]
        yi = xyzi_3[i,1]
        zi = xyzi_3[i,2]
        # 计算出每个点云所属的体素位置
        x_id = round(np.floor((xi-min_p[0])/density_xy))
        y_id = round(np.floor((yi-min_p[1])/density_xy))
        z_id = round(np.floor((zi-min_p[2])/density_xy))
        voxel[x_id, y_id, z_id, 3] += 1  # 点云计数器
        voxel[x_id, y_id, z_id, 0] += xyzi_3[i, 3]
        voxel[x_id, y_id, z_id, 1] += xyzi_3[i, 4]
        voxel[x_id, y_id, z_id, 2] += xyzi_3[i, 5]  # 强度值求和
        # 体元填值
        LWH[i, 0] = x_id
        LWH[i, 1] = y_id
        LWH[i, 2] = z_id
    # 强度值求平均
    voxel[:, :, :, 0] /= voxel[:, :, :, 3]
    voxel[:, :, :, 1] /= voxel[:, :, :, 3]
    voxel[:, :, :, 2] /= voxel[:, :, :, 3]
    voxel = np.nan_to_num(voxel)
    # 体素标准化
    if S==1:
        voxel[:, :, :, 0] =voxel[:, :, :, 0] * voxel[:, :, :, 1] * voxel[:, :, :, 2]
        voxel = voxel[:,:,:,0]

    return voxel, LWH

# 体素显示
def voxel_view(voxel):
    fig = plt.figure()  # 新建一个窗口
    ax = fig.gca(projection='3d')  # 使用三维显示
    ax.voxels(voxel, edgecolor="k")  # 连接体素容器  （需要改进一下显示颜色）
    plt.show()  # 显示窗口

# 高斯混合模型(输入点云分类特征、总的类别数)加上显示
def use_gmm(data,classes):
    # 显示初始条件的分布散点图
    if np.shape(data)[-1] == 2:
        plt.figure()
        plt.scatter(data[:, 0],data[:, 1])
        plt.show()
    elif np.shape(data)[-1] == 3:
        fig = plt.figure()
        ax = Axes3D(fig)
        ax.scatter(data[:, 0],data[:, 1],data[:, 2])
        plt.show()
    # 进行高斯混合模型分解
    gmm = GaussianMixture(classes)
    gmm.fit(data)
    labels = gmm.predict(data)  # 判断每个点云所属的类别
    # 显示分类后的分布散点图（无法保证类别，暂时忽略）
    return labels

# 求点云的平面密度
def num_area(num,area):
    return num/area

# 求RMSE
def get_RMSE(data):
    sum_d = sum(data)
    n = len(data)
    return sum_d/n

# 建立断点
def cut_down(num, Piece=mp.cpu_count()):
    tik = []
    if num <= Piece:
        tik.append(0)
        print('点云数量过少，不能分块')
    else:
        n_pool = math.ceil(num / Piece)  # 每个池处理的最大点云数量
        print('每个block的tik位置为', n_pool)
        for i in range(0, Piece):
            tik.append(i * n_pool)
    tik.append(num)
    return tik  # 输出每个断点位置

# 并行计算每个点的反正切
def find_arctan_block(xzc_point, xzc_c, tik0, tik1):
    d_yx_ = np.empty([tik1 - tik0, 2])
    j = 0  # 计数器
    for i in range(tik0, tik1):
        ci = xzc_point[i, -1]
        xz_c = xzc_c[xzc_c[:, -1] == ci, :2]
        d_yx_[j, 0] = (xzc_point[i, 1] - xz_c[0, 1])
        d_yx_[j, 1] = (xzc_point[i, 0] - xz_c[0, 0])
        j += 1
    print('已完成', tik0, '-', tik1, '的反正切')
    return d_yx_

# 计算每个点云属于哪个环块
def find_belong_block(angle, single_a, tik0, tik1):
    belong_ = np.empty(tik1 - tik0)
    j = 0  # 计数器
    for i in range(tik0, tik1):
        belong_[j] = np.floor(angle[i] / single_a)  # 点云归属标签赋值
        j += 1
    print('已完成', tik0, '-', tik1, '的环块归属')
    return belong_

# 将圆环分成n份(点云以及圆环)并行计算
def Split_ring_mp(xzc_point, xzc_c, n=10160, num_cpu=mp.cpu_count()):
    # 求每个点的反正切
    num = len(xzc_point)  # 点云数量
    d_yx = np.empty([num, 2])  # 存储点云截距的容器
    belong = np.empty([num])  # 存储点云标签值容器
    # 并行计算准备
    tik = cut_down(num, num_cpu)  # 分块起止点
    pool = mp.Pool(processes=num_cpu)  # 开启多进程池，数量为cpu
    multi_res = [pool.apply_async(find_arctan_block, args=(xzc_point, xzc_c, tik[i], tik[i + 1])) for i in
                 range(num_cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    tik_ = 0  # 计数器
    for res in multi_res:
        d_yx[tik[tik_]:tik[tik_ + 1], :] = res.get()
        tik_ += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    angle = np.arctan2(d_yx[:, 0], d_yx[:, 1])  # 求反正切值
    # 将角度分解
    single_a = np.pi * 2 / n  # 求每个块的过渡
    # 并行计算点云归属块
    pool = mp.Pool(processes=num_cpu)  # 开启多进程池，数量为cpu
    multi_res = [pool.apply_async(find_belong_block, args=(angle, single_a, tik[i], tik[i + 1])) for i in
                 range(num_cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    tik_ = 0  # 计数器
    for res in multi_res:
        belong[tik[tik_]:tik[tik_ + 1]] = res.get()
        tik_ += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    return belong

def get_ρθ_block(xyzic, xzrc, Ts, Te, n):
    '分块求圆心距和角度函数'
    # 准备工作
    single_a = np.pi * 2 / n  # 求每个块的过渡
    ρθ_ = np.empty([Te-Ts, 2])
    j = 0  # 计数器
    for i in range(Ts, Te):
        # 求p
        xyz_ = xyzic[i, :3]  # 当前点的空间位置
        c_i = xyzic[i, -1]  # 当前点圆环名
        xzr_ = xzrc[xzrc[:, -1] == c_i, :3]  # 圆心位置
        R2 = (xyz_[0] - xzr_[0, 0]) ** 2 + (xyz_[2] - xzr_[0, 1]) ** 2  # 点到圆心的距离
        ρθ_[j, 0] = np.sqrt(R2)  # 点云到中轴线的距离
        # 求角度
        angle_ = np.arctan2((xyzic[i, 2] - xzr_[0, 1]), (xyzic[i, 0] - xzr_[0, 0]))  # 求反正切值
        ρθ_[j, 1] = np.floor(angle_ / single_a)  # 点云归属标签赋值
        # 其他
        j += 1
    print('极坐标系转换已完成', Te/len(xyzic)*100, '%')
    return ρθ_

def get_ρθ_block2(xyzic, xyzc, Ts, Te, n):
    '分块只求ρθ的函数'
    # 准备工作
    single_a = np.pi * 2 / n  # 求每个块的过渡
    ρθ_ = np.empty([Te - Ts, 2])
    j = 0  # 计数器
    for i in range(Ts, Te):
        # 求圆心距
        xyz_ = xyzic[i, :3]  # 当前点的空间位置
        c_i = xyzic[i, -1]  # 当前点圆环名
        c_xyz_ = xyzc[xyzc[:, -1]==c_i, :3]  # 圆心位置
        R2 = (xyz_[0] - c_xyz_[0, 0]) ** 2 + (xyz_[2] - c_xyz_[0, 2]) ** 2  # 点到圆心的距离
        ρθ_[j, 0] = np.sqrt(R2)  # 点云到中轴线的距离
        # 求角度
        angle_ = np.arctan2((xyzic[i, 2] - c_xyz_[0, 2]), (xyzic[i, 0] - c_xyz_[0, 0]))  # 求反正切值
        ρθ_[j, 1] = np.floor(angle_ / single_a)  # 点云归属标签赋值
        # ρθ_[j, 1] = angle_
        j += 1  # 计数器更新
    print('极坐标属性添加已完成', Te / len(xyzic) * 100, '%')
    return ρθ_

def XYZ2ρθ_mp(xyzic,xyzc,num_cpu=mp.cpu_count(),n=4250):
    '中轴线刷新后的并行计算 笛卡尔坐标系转成圆柱体坐标系(非体素化准备)'
    num = len(xyzic)  # 点云数量
    tik = cut_down(num)  # 并行计算分块
    ρθ = np.empty([num, 2])  # 新建一个存储点云ρθ的容器
    pool = mp.Pool(processes=num_cpu)  # 开启多进程池，数量为cpu
    # 开始进行并行计算
    multi_res = [pool.apply_async(get_ρθ_block2, args=(xyzic, xyzc, tik[i], tik[i + 1], n)) for i in
                 range(num_cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    tik_ = 0  # 计数器
    for res in multi_res:
        ρθ[tik[tik_]:tik[tik_ + 1], :] = res.get()
        tik_ += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    # print(np.max(ρθ[:, 1]),np.min(ρθ[:, 1]))
    # 归零化
    min_ρθ = ρθ.min(0)  # 求最小值
    ρθ[:, 0] -= min_ρθ[0]
    ρθ[:, 1] -= min_ρθ[1]
    return ρθ

def XYZ2ρθY_mp(xyzic,xzrc,num_cpu=mp.cpu_count(),n=3000):
    '并行计算 笛卡尔坐标系转成圆柱体坐标系（体素化准备）'
    # 准备工作
    num = len(xyzic)  # 点云数量
    tik = cut_down(num)  # 并行计算分块
    ρθY = np.empty([num, 3])  # 新建一个存储点云ρθ的容器
    ρθY[:, -1] = xyzic[:, 1]  # 传Y
    pool = mp.Pool(processes=num_cpu)  # 开启多进程池，数量为cpu
    # 开始进行并行计算
    multi_res = [pool.apply_async(get_ρθ_block, args=(xyzic, xzrc, tik[i], tik[i + 1], n)) for i in
                 range(num_cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    tik_ = 0  # 计数器
    for res in multi_res:
        ρθY[tik[tik_]:tik[tik_ + 1], :2] = res.get()
        tik_ += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    # 归0化
    min_ρθY = ρθY.min(0)  # 求最小值
    ρθY[:,0] -= min_ρθY[0]
    ρθY[:, 1] -= min_ρθY[1]
    ρθY[:, 2] -= min_ρθY[2]  # 三列归零
    return ρθY

def XYZ2ρθY(xyzic,xzrc,n=3000):
    '笛卡尔坐标系转成圆柱体坐标系'
    Y = xyzic[:, 1]  # y轴坐标为起点
    c = xzrc[:, -1]  # 所有圆环名
    ρ = np.empty([len(xyzic)])  # 新建一个存储点云p的容器
    '准备添加并行化'
    for i in range(len(xyzic)):
        xyz_ = xyzic[i, :3]  # 当前点的空间位置
        c_i = xyzic[i, -1]  # 当前点圆环名
        xzr_ = xzrc[c == c_i, :3]  # 圆心位置
        # print(i,xyz_,xzr_)
        R2 = (xyz_[0] - xzr_[0, 0]) ** 2 + (xyz_[2] - xzr_[0, 1]) ** 2  # 点到圆心的距离
        # dis[i] = np.abs(R2 - xzr_[0, -1] ** 2)
        ρ[i] = np.sqrt(R2)  # 点云到中轴线的距离

    θ = np.empty([len(xyzic)])  # 新建一个存储点云θ的容器
    # 求角度
    xzc_point = np.c_[xyzic[:,0],xyzic[:,2],xyzic[:,-1]]
    xzc_c = np.c_[xzrc[:,:2],xzrc[:,-1]]
    θ = Split_ring_mp(xzc_point,xzc_c,n=n)
    ρθY = np.c_[ρ,θ,Y]
    # 归一化
    θ_min = np.min(ρθY[:, 1])  # 角度向量化后的最小值
    ρθY[:, 1] = ρθY[:, 1] - θ_min  # 使所有的角度都大于0
    Y_min = np.min(ρθY[:, -1])  # Y轴后的最小值
    ρθY[:, -1] = ρθY[:, -1] - Y_min  # Y轴归0
    return ρθY

def FitGaussian1D(data,bins=20):
    '''
    单高斯模型拟合
    :param data: 需要拟合的数据（一维）
    :param bins: 按照data的值域分成多少份，默认为20份
    :return:
    Fx.mean.value：拟合后高斯分布的均值；
    Fx.stddev.value：拟合后高斯分布的标准差
    '''
    y_hist,x_hist = np.histogram(data, bins)  # 使用numpy粗拟合，得到直方图曲线点
    x_hist = (x_hist[1:] + x_hist[:-1]) / 2  # 刷新值域断点
    model_Gaussian1D = models.Gaussian1D(amplitude=np.max(y_hist), mean=np.mean(data), stddev=np.std(data))  # 建立高斯模型
    model_Fit = fitting.LevMarLSQFitter()  # 建立拟合模型
    Fx = model_Fit(model_Gaussian1D, x_hist, y_hist)  # 模型求解
    '''
    x = np.linspace(0, 255, int(255/bins))
    y = Fx(x)
    plt.plot(x, y)
    plt.show()
    '''
    return Fx.mean.value,Fx.stddev.value

def get_distance_point2line(points, line_ab):
    """
    点到直线的距离（平面斜率截距式）
    Args:
        points: [x0, y0]*n
        line_ab: [k, b]
    """
    k, b = line_ab
    distance = abs(k * points[:,0] - points[:,1] + b) / math.sqrt(k**2 + 1)
    # distance = abs(k * point[0] - point[1] + b) / math.sqrt(k ** 2 + 1)
    return distance



import os
import numpy as np
import zelas2.TheHeartOfTheMilitaryGod as zt
import random
import mmengine
import zelas2.shield as zs

def get_pts_paths(folder_path, end='las'):
    """
    读取当前文件夹下所有点云数据的绝对路径并保存到一个数组中

    Args:
        folder_path (str): 文件夹路径

    Returns:
        list: 所有las点云数据的绝对路径列表
    """
    las_paths = []
    for filename in os.listdir(folder_path):
        if filename.endswith(end):
            print(os.path.join(folder_path, filename))
            las_paths.append(os.path.join(folder_path, filename))
    return las_paths

def cut_points(xyzi,output_dir,name,chunk_size=160,old_max_i=65535):
    '将大点云分割成小点云,并重置xy坐标系'
    min_x = np.min(xyzi[:, 0])
    max_x = np.max(xyzi[:, 0])
    min_y = np.min(xyzi[:, 1])
    max_y = np.max(xyzi[:, 1])

    num_chunks_x = int(np.ceil((max_x - min_x) / chunk_size))
    num_chunks_y = int(np.ceil((max_y - min_y) / chunk_size))

    for i in range(num_chunks_x):
        for j in range(num_chunks_y):
            chunk_min_x = min_x + i * chunk_size
            chunk_max_x = chunk_min_x + chunk_size
            chunk_min_y = min_y + j * chunk_size
            chunk_max_y = chunk_min_y + chunk_size

            xyzi_ = xyzi[xyzi[:,0]>=chunk_min_x,:]
            xyzi_ = xyzi_[xyzi_[:,0]<chunk_max_x,:]
            xyzi_ = xyzi_[xyzi_[:,1]>=chunk_min_y,:]
            xyzi_ = xyzi_[xyzi_[:,1]<chunk_max_y,:]
            '重置坐标系'
            center_x = (chunk_min_x + chunk_max_x) / 2
            center_y = (chunk_min_y + chunk_max_y) / 2

            xyzi_[:, 0] = xyzi_[:, 0] - center_x
            xyzi_[:, 1] = xyzi_[:, 1] - center_y
            xyzi_[:, 3] = xyzi_[:, 3] / old_max_i
            output_path = output_dir+'\\'+name+str(i)+str(j)+'.txt'
            np.savetxt(output_path,xyzi_,fmt='%.05f')
            print('已经输出',output_path,'点云')

def cut_points_NS(xyzi,output_dir,name,chunk_size=125,old_max_i=65535):
    '将大点云分割成小点云,并重置xy坐标系,左下角为原点'
    # 求原始点云四至
    min_x = np.min(xyzi[:, 0])
    max_x = np.max(xyzi[:, 0])
    min_y = np.min(xyzi[:, 1])
    max_y = np.max(xyzi[:, 1])
    # 求分割边长
    num_chunks_x = int(np.ceil((max_x - min_x) / chunk_size))
    num_chunks_y = int(np.ceil((max_y - min_y) / chunk_size))
    # 分割坐标系+重置坐标系
    for i in range(num_chunks_x):
        for j in range(num_chunks_y):
            # 每块数据的四至
            chunk_min_x = min_x + i * chunk_size
            chunk_max_x = chunk_min_x + chunk_size
            chunk_min_y = min_y + j * chunk_size
            chunk_max_y = chunk_min_y + chunk_size
            # 找到这块数据
            xyzi_ = xyzi[xyzi[:,0]>=chunk_min_x,:]
            xyzi_ = xyzi_[xyzi_[:,0]<chunk_max_x,:]
            xyzi_ = xyzi_[xyzi_[:,1]>=chunk_min_y,:]
            xyzi_ = xyzi_[xyzi_[:,1]<chunk_max_y,:]
            '重置坐标系'
            # center_x = (chunk_min_x + chunk_max_x) / 2
            # center_y = (chunk_min_y + chunk_max_y) / 2
            xyzi_[:, 0] = xyzi_[:, 0] - chunk_min_x
            xyzi_[:, 1] = xyzi_[:, 1] - chunk_min_y
            xyzi_[:, 3] = xyzi_[:, 3] / old_max_i
            # 输出点云
            output_path = output_dir+'\\'+name+str(i)+str(j)+'.txt'
            np.savetxt(output_path,xyzi_,fmt='%.05f')
            print('已经输出',output_path,'点云')


def reset_z_intensity(xyzil,old_max_i=65535):
    '重置z坐标和强度值'
    xyzil[:,3] = xyzil[:,3]/old_max_i  # 重置强度值

    center_z = np.min(xyzil[:,2])  # 获取最低高程值
    xyzil[:,2] = xyzil[:,2] - center_z  # 矫正高程值，使得都大于0
    return xyzil

def reset_y(xyzil):
    '重置y坐标'
    y_mid = np.min(xyzil[:,1])+np.ptp(xyzil[:,1])/2
    xyzil[:,1] = xyzil[:,1] - y_mid
    return xyzil

def reset_newXYZI(xyzil,old_max_i=255):
    '重置圆柱体3D坐标以及强度值'
    z0 = np.mean(xyzil[:,2])  # 求z中心
    xyzil[:, 3] = xyzil[:, 3] / old_max_i  # 重置强度值
    # 拟合圆中心
    model = zs.CircleLeastSquareModel()  # 类的实例化:用最小二乘生成已知模型
    data = np.vstack([xyzil[:, 0], xyzil[:, 1]]).T  # 整理数据
    result = model.fit(data)  # 拟合圆
    x0 = result.a * -0.5
    y0 = result.b * -0.5
    # 刷新坐标
    print('新圆心',x0,y0,z0)
    xyzil[:, 0] = xyzil[:, 0] - x0
    xyzil[:, 1] = xyzil[:, 1] - y0
    xyzil[:, 2] = xyzil[:, 2] - z0
    
    return xyzil

def np2bin_batch(input,otput):
    'numpy点云批量转为.bin格式'
    # 获取点云路径
    files = [f for f in os.listdir(input) if f.endswith('.txt')]
    # 批量转为.bin
    for file in files:
        xyzil_ = zt.ReadByPolars(os.path.join(input, file))
        xyzi_ = xyzil_[:,:4]
        xyzi_ = np.float32(xyzi_)
        print('已经读取点云',os.path.join(input, file))
        name_ = os.path.splitext(file)[0] + '.bin'
        with open(os.path.join(otput,name_), 'wb') as f:
            f.write(xyzi_.tobytes())
        print('已经输出点云',os.path.join(otput, name_))

def np2label_batch(input,otput):
    'numpy点云标签批量生成.bin格式标签'
    files = [f for f in os.listdir(input) if f.endswith('.txt')]
    '标签批量转为.bin'
    for file in files:
        xyzil_ = zt.ReadByPolars(os.path.join(input, file))
        print('已经读取点云',os.path.join(input, file))
        L = np.int32(xyzil_[:,4])
        print(np.unique(L))
        name_ = os.path.splitext(file)[0] + '.label'
        with open(os.path.join(otput,name_), 'wb') as f:
            f.write(L.tobytes())
        print('已经输出逐点级标签',os.path.join(otput, name_))

def np2ImageSets_batch(input_path,train_file,val_file,test_file,p=np.array([7,2,1])):
    '生成训练集、验证集和测试集.txt'
    '批量读取las'
    files = [f for f in os.listdir(input_path) if f.endswith('.bin')]
    random.shuffle(files)  # 打乱数据集
    p_sum = np.sum(p)
    total_length = len(files)
    part1_length = np.floor(total_length * p[0] / p_sum).astype(int)
    part2_length = np.floor(total_length * p[1] / p_sum).astype(int)
    # part3_length = total_length - part1_length - part2_length
    part1 = files[:part1_length]
    part2 = files[part1_length:part1_length + part2_length]
    part3 = files[part1_length + part2_length:]
    # 将训练集和验证集\测试集的文件名写入对应的 .txt 文件
    with open(train_file, 'w') as train_f:
        for file in part1:
            train_f.write(f"{file}\n")
    with open(val_file, 'w') as val_f:
        for file in part2:
            val_f.write(f"{file}\n")
    with open(test_file, 'w') as test_f:
        for file in part3:
            test_f.write(f"{file}\n")

def create_pkl(ImageSets_path,out_file,data_set='Custom'):
    '通过数据集生成mmdet3d所能接受的仿seg_kitti的.pkl'
    data_infos = dict()
    data_infos['metainfo'] = dict(DATASET=data_set)
    data_list = []
    # 读取ImageSets里面的txt
    with open(ImageSets_path, 'r') as f:
        # 读取所有行并去除每行的换行符
        samples = [line.strip() for line in f.readlines()]
    num_samples = len(samples)
    for i in range(num_samples):
        data_list.append({
            'lidar_points': {
                'lidar_path':
                    'points'+'/'+samples[i],
                'num_pts_feats':
                    4
            },
            'pts_semantic_mask_path':
                'semantic_mask'+'/'+os.path.splitext(samples[i])[0] + '.label',
            'sample_id':
                str(i)
        })
    data_infos.update(dict(data_list=data_list))
    # 保存.pkl
    mmengine.dump(data_infos, out_file)


import ezdxf
from ezdxf import colors
from tqdm import tqdm
from ezdxf.addons import odafc
import numpy as np

def normalize_color(value):
    'RGB unit16转8'
    return int((value / 65535) * 255)


def points2dxf(xyzrgb,output=None,dxfversion='R2013'):
    '''
    点云转dxf
    :param dxfversion: AutoCAD版本号
    :param xyzrgb: 记录每个点云的坐标和颜色
    :param output: 输出路径
    :return: dxf类
    '''
    num = len(xyzrgb)  # 点云数量
    doc = ezdxf.new(dxfversion=dxfversion)  # 新建一个DXF类
    msp = doc.modelspace()  # DXF模型指针
    # 设置新的边界点位置
    extmin = (np.min(xyzrgb[:,0]), np.min(xyzrgb[:,1]), np.min(xyzrgb[:,2]))  # 最小边界
    extmax = (np.max(xyzrgb[:,0]), np.max(xyzrgb[:,1]), np.max(xyzrgb[:,2]))  # 最大边界
    msp.reset_extents(extmin, extmax)  # 重置边界
    # 遍历las点云添加.dxf点位置及颜色（真彩色）
    for i in tqdm(range(num)):
        color = colors.rgb2int((normalize_color(xyzrgb[i,3]), normalize_color(xyzrgb[i,4]), normalize_color(xyzrgb[i,5])))  # 将颜色转换为 AutoCAD 颜色
        # 将点添加到 DXF 文件中
        dxf_point = msp.add_point((xyzrgb[:,0], xyzrgb[:,1], xyzrgb[:,2]))
        # 设置附加属性
        dxf_point.dxf.linetype = 'Continuous'  # 点的线型
        dxf_point.dxf.color = 256  # AutoCAD 颜色号  # ['62']
        dxf_point.dxf.true_color = color  # 点的真彩颜色
        dxf_point.dxf.lineweight = 13  # 线宽 单位毫米
    # 保存
    if output is not None:
        doc.saveas(output)  # 保存.dxf
    return doc


def dxf2dwg(doc,output,version='R13',by='C:\\Program Files\\ODA\\ODAFileConverter 25.4.0\\ODAFileConverter.exe'):
    '''
    dxf转dwg
    :param doc: dxf文件类
    :param output: dwg输出路径
    :param version: dwg版本号
    :param by: ODA.exe路径
    '''
    odafc.win_exec_path = by  # 设置软件开源路径
    odafc.export_dwg(doc, output, replace=True, version=version)  # 保存.dwg



# 本代码适用于25年3月的纵缝提取
from typing import Union
import numpy as np
import zelas2.shield as zs
import multiprocessing as mp
import zelas2.Multispectral as zm
from sympy.codegen.ast import Return
from tqdm import tqdm
import cv2
from scipy.spatial.transform import Rotation
import zelas2.RedundancyElimination as zr
from sklearn.neighbors import KDTree  # 添加机器学习的skl.KDT的函数组
from multiprocessing import shared_memory
from sklearn.cluster import DBSCAN
import zelas2.ransac as zR
import zelas2.TheHeartOfTheMilitaryGod as zt
from scipy.spatial.distance import pdist, squareform
from scipy.signal import find_peaks

def find_continuous_segments_numpy(arr):
    """
    找到一维数组中所有连续整数段的起始和终止数，返回 NumPy 数组
    :param arr: 一维整数数组（已排序）
    :return: NumPy 数组，每行为 (起始数, 终止数)
    """
    segments = []  # 存储所有连续段
    start = arr[0]  # 当前连续段的起始数
    for i in range(1, len(arr)):
        if arr[i] != arr[i - 1] + 1:  # 检测中断点
            segments.append((start, arr[i - 1]))  # 保存当前连续段
            start = arr[i]  # 开始新的连续段
    # 添加最后一个连续段
    segments.append((start, arr[-1]))
    # 转换为 NumPy 数组
    return np.array(segments, dtype=int)

def get_ρθ(xz_p, xzr):
    '求每个盾构环的极径差和反正切'
    num_p = len(xz_p)  # 当前截面点数量
    ρ = np.sqrt((xz_p[:,0]-xzr[0])**2+(xz_p[:,1]-xzr[1])**2)-xzr[2]
    θ = np.empty(num_p)
    for i in range(num_p):
        θ[i] = zs.get_angle(xz_p[i,0],xz_p[i,1],xzr[0],xzr[1])
    return np.c_[ρ,θ]

def find_seed(θyρvci,ρ_td,r,cpu=mp.cpu_count(),c_ignore=4):
    '寻找符合纵缝特征的种子点'
    θyρvci_up = θyρvci[θyρvci[:,2]>=ρ_td,:]  # 低于衬砌点的不要
    # 准备工作
    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    c_un = np.unique(θyρvci[:,4])
    num_c = len(np.unique(θyρvci[:,4]))  # 截面数
    # good_index = []
    # 并行计算
    multi_res = pool.starmap_async(find_seed_cs, ((θyρvci,np.uint64(θyρvci_up[θyρvci_up[:,4]==c_un[i],5]),c_un[i],r,c_ignore) for i in
                 tqdm(range(num_c),desc='分配任务寻找种子点',unit='个截面',total=num_c)))
    j = 0
    for res in tqdm(multi_res.get(),total=num_c,desc='输出种子点下标'):
        if j==0:
            good_index = res
        else:
            good_index = np.hstack((good_index, res))
        j += 1
    pool.close()  # 关闭所有进程
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    return np.int64(good_index)


def find_seed_cs(θyρvci,id_θyρvci_up_c,c,r,c_ignore=4):
    '''
    寻找单个截面符合纵缝特征的种子点
    θyρvci : 点云信息
    id_θyρvci_up_c ：符合搜索的点云下标
    c ：当前截面
    r :当前截面半径
    c_ignore ：忽略的截面数
    '''
    good_ind = []  # 空下标
    θ_l_td = 0.15 * np.pi * r / 180
    for i in id_θyρvci_up_c:
        if θyρvci[i,3]==0:  # 如果当前点为线特征
            # 找到搜索截面
            θyρ_c_ = θyρvci[θyρvci[:, 4] <= c + c_ignore, :]
            θyρ_c_ = θyρ_c_[θyρ_c_[:, 4] >= c - c_ignore, :]
            # 判断左侧是否有球特征
            θ_l_ = θyρvci[i,0] - θ_l_td  # 左侧角度阈值
            θyρ_l_ = θyρ_c_[θyρ_c_[:, 0] < θyρvci[i, 0], :]
            θyρ_l_ = θyρ_l_[θyρ_l_[:, 0] >= θ_l_, :]
            # 判断右侧是否有球特征
            θ_r_ = θyρvci[i,0] + θ_l_td  # 右侧角度阈值
            θyρ_r_ = θyρ_c_[θyρ_c_[:, 0] > θyρvci[i, 0], :]
            θyρ_r_ = θyρ_r_[θyρ_r_[:, 0] <= θ_r_, :]
            if 2 in θyρ_l_[:, 3] and 2 in θyρ_r_[:, 3]:  # 如果左右都有球特征
                good_ind.append(θyρvci[i,5])  # 作为种子点
    return good_ind

def distance_to_line(point, line):
    """计算点到直线的几何距离"""
    x0, y0 = point
    x1, y1, x2, y2 = line
    numerator = abs((y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1)
    denominator = np.sqrt((y2 - y1) ** 2 + (x2 - x1) ** 2)
    return numerator / denominator if denominator != 0 else 0

def merge_similar_lines(lines, angle_thresh=np.pi / 18, rho_thresh=20):
    """合并相似直线（极坐标参数相近的线段）"""
    merged = []
    for line in lines:
        rho, theta = line_to_polar(line[0])
        found = False
        for m in merged:
            m_rho, m_theta = m[0]
            # 检查角度和距离差异
            if abs(theta - m_theta) < angle_thresh and abs(rho - m_rho) < rho_thresh:
                m[0] = ((m_rho + rho) / 2, (m_theta + theta) / 2)  # 合并平均值
                m[1].append(line)
                found = True
                break
        if not found:
            merged.append([(rho, theta), [line]])

    # 转换回线段格式（取合并后的极坐标生成新线段）
    merged_lines = []
    for m in merged:
        rho, theta = m[0]
        a = np.cos(theta)
        b = np.sin(theta)
        x0 = a * rho
        y0 = b * rho
        # 生成足够长的线段（覆盖图像范围）
        scale = 1000
        x1 = int(x0 + scale * (-b))
        y1 = int(y0 + scale * (a))
        x2 = int(x0 - scale * (-b))
        y2 = int(y0 - scale * (a))
        merged_lines.append([x1, y1, x2, y2])

    return merged_lines[:5]  # 最多返回前5条

def find_lines(θy):
    '通过数字图像操作将纵缝找到并返回种子点'
    '0.整理数据'
    θy = θy*100
    θy[:, 0] -= np.min(θy[:, 0])
    θy[:, 1] -= np.min(θy[:, 1])
    θy = np.uint64(θy)
    x_max = np.max(θy[:,0])
    y_max = np.max(θy[:,1]) # 求边界
    print('二维边长',x_max,y_max)
    '1.创建图像'
    img = np.zeros((int(y_max)+1, int(x_max)+1), dtype=np.uint8)
    for x,y in θy:
        img[y, x] = 255
    '''
    cv2.imshow("Image", img)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    '''
    '2.直线检测'
    # 霍夫变换检测直线
    lines = cv2.HoughLinesP(img, rho=1, theta=np.pi / 180, threshold=50, minLineLength=50, maxLineGap=50)  # 检测单位像素、角度、超过像素阈值、最小长度阈值、最大间断阈值
    # --- 提取前5条最长的直线 ---
    detected_lines = []
    if lines is not None:
        lines = lines[:, 0, :]
        # 按线段长度排序（从最长到最短）
        lines = sorted(lines, key=lambda x: np.linalg.norm(x[2:] - x[:2]), reverse=True)[:6]
        detected_lines = lines
    threshold_distance = 2.0  # 点到直线的最大允许距离（根据噪声调整）
    # 初始化：所有点标记为未分配
    assigned = np.zeros(len(θy), dtype=bool)
    line_points_list = []  # 存储每条直线的点
    line_indices_list = []  # 存储每条直线的点索引
    for line in detected_lines:
        distances = np.array([distance_to_line(p, line) for p in θy])
        # 筛选未分配且距离小于阈值的点
        mask = (distances < threshold_distance) & ~assigned
        line_points = θy[mask]
        line_points_list.append(line_points)
        indices = np.where(mask)[0]
        line_indices_list.append(indices)
        assigned |= mask  # 标记已分配的点
    # 合并前5条直线的点
    all_line_points = np.vstack(line_points_list)
    # 分离噪声点
    noise_points = θy[~assigned]
    '''
    # --- 可视化结果 ---
    # 创建彩色图像用于显示
    result_img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    # 绘制检测到的直线（绿色）
    for line in detected_lines:
        x1, y1, x2, y2 = line
        cv2.line(result_img, (x1, y1), (x2, y2), (0, 255, 0), 2)
    # 绘制属于直线的点（红色）
    for p in all_line_points:
        cv2.circle(result_img, tuple(p), 2, (0, 0, 255), -1)
    '''
    '''
        # 预定义5种颜色（BGR格式）
    colors = [
        (0, 0, 255),   # 红色
        (0, 255, 0),   # 绿色
        (255, 0, 0),   # 蓝色
        (0, 255, 255), # 黄色
        (255, 0, 255)  # 品红色
    ]
    # 绘制每条直线及其对应的点
    for i, (line, line_points) in enumerate(zip(detected_lines, line_points_list)):
        color = colors[i % len(colors)]  # 循环使用颜色列表
        # 绘制直线
        x1, y1, x2, y2 = line
        cv2.line(result_img, (x1, y1), (x2, y2), color, 2)
    '''
    '''
    # 绘制噪声点（蓝色）
    for p in noise_points:
        cv2.circle(result_img, tuple(p.astype(int)), 2, (255, 0, 0), -1)
    cv2.imshow("Result", result_img)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    '''
    # 去除 line_indices_list 中的空元素
    line_indices_list = [indices for indices in line_indices_list if len(indices) > 0]
    return line_indices_list


def merge_similar_lines(lines, angle_thresh=5, dist_thresh=10):
    """
    合并角度和位置相近的线段
    :param lines: 线段列表，格式为 [[x1,y1,x2,y2], ...]
    :param angle_thresh: 角度差阈值（度）
    :param dist_thresh: 线段中心点距离阈值（像素）
    :return: 合并后的线段列表
    """
    merged = []
    for line in lines:
        x1, y1, x2, y2 = line
        # 计算线段角度（弧度）
        angle = np.arctan2(y2 - y1, x2 - x1) * 180 / np.pi
        # 计算线段中心点
        cx, cy = (x1 + x2) / 2, (y1 + y2) / 2

        # 检查是否与已合并线段近似
        found = False
        for m in merged:
            m_angle, m_cx, m_cy = m['angle'], m['cx'], m['cy']
            # 角度差和中心点距离
            angle_diff = abs(angle - m_angle)
            dist = np.sqrt((cx - m_cx) ** 2 + (cy - m_cy) ** 2)

            if angle_diff < angle_thresh and dist < dist_thresh:
                # 合并线段（延长端点）
                m['x1'] = min(m['x1'], x1, x2)
                m['y1'] = min(m['y1'], y1, y2)
                m['x2'] = max(m['x2'], x1, x2)
                m['y2'] = max(m['y2'], y1, y2)
                found = True
                break

        if not found:
            merged.append({
                'x1': x1, 'y1': y1, 'x2': x2, 'y2': y2,
                'angle': angle, 'cx': cx, 'cy': cy
            })
    # 转换为坐标格式
    return [[m['x1'], m['y1'], m['x2'], m['y2']] for m in merged]


def fit_3d_line(points):
    """
    Fit a 3D line to a point cloud using PCA.
    Parameters:
    points (numpy.ndarray): Nx3 array of 3D points.
    Returns:
    tuple: (centroid, direction_vector)
        centroid is a point on the line (numpy.ndarray of shape (3,)),
        direction_vector is the direction vector of the line (numpy.ndarray of shape (3,)).
    """
    # 计算点云的质心
    centroid = np.mean(points, axis=0)
    # 将点云中心化
    centered_points = points - centroid
    # 计算协方差矩阵
    cov_matrix = np.cov(centered_points.T)
    # 计算协方差矩阵的特征值和特征向量
    eigenvalues, eigenvectors = np.linalg.eigh(cov_matrix)
    # 找到最大特征值对应的特征向量作为方向向量
    direction_vector = eigenvectors[:, np.argmax(eigenvalues)]
    # Centroid (a point on the line): [1.5 1.5 1.5]
    # Direction vector: [0.57735027 0.57735027 0.57735027]
    return centroid, direction_vector


def distance_to_line_3D(points, centroid, direction):
    """
    计算点到三维直线的距离。
    Parameters:
    points (numpy.ndarray): Nx3 的3D点。
    centroid (numpy.ndarray): 直线上的一点，形状为 (3,)。
    direction (numpy.ndarray): 直线的单位方向向量，形状为 (3,)。
    Returns:
    numpy.ndarray: 每个点到直线的距离，形状为 (N,)。
    """
    # 计算点与质心的向量差
    vec = points - centroid
    # 计算叉乘 (支持批量计算)
    cross_product = np.cross(vec, direction)
    # 距离为叉乘的模长
    distances = np.linalg.norm(cross_product, axis=1)
    return distances

def find_CS_25(xyzic,GirthInterval=245,num_cpu=mp.cpu_count(),z_range=1):
    '环缝提取，固定长度版，25年修补版'
    xyzic[:, 3] = zm.normalization(xyzic[:, 3], 255)  # 强度值离散化
    # 计算每个环的平均强度值
    c_un = np.unique(xyzic[:, 4])  # 圆环从小到大排列
    num_C = len(c_un)  # 截面数量
    # 并行计算准备
    tik = zs.cut_down(num_C, num_cpu)  # 分块起止点
    pool = mp.Pool(processes=num_cpu)  # 开启多进程池，数量为cpu
    # 限制参与计算的比例
    z_max = np.max(xyzic[:, 2])
    z_min = np.min(xyzic[:, 2])
    d_z = z_max - z_min
    t_z = z_min + d_z * (1 - z_range)  # 参与统计的阈值
    ps_free = xyzic[xyzic[:, 2] >= t_z, :]  # 参与统计的点云
    i_c = np.empty(num_C)  # 新建一个存储圆环平均强度值的容器
    multi_res = [pool.apply_async(zs.find_cImean_block, args=(ps_free, c_un, tik[i], tik[i + 1])) for i in
                 range(num_cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    tik_ = 0  # 计数器
    for res in multi_res:
        i_c[tik[tik_]:tik[tik_ + 1]] = res.get()
        tik_ += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    # 后期整理
    i_c = zm.normalization(i_c, 255)  # 均值离散化
    i_c_mean = np.mean(i_c)  # 平均强度值均值
    i_c_std = np.std(i_c)  # 平均强度值方差
    print('截面强度值均值', i_c_mean, '截面强度值标准差', i_c_std)
    # 找到强度值的最低点以及其他的极低点
    id_min = np.argmin(i_c)
    num_CS_0 = int(np.round(num_C / GirthInterval))  # 假想环缝数量
    print('理论的环缝数量为',str(num_CS_0))
    Begin_CS_0 = id_min % GirthInterval  # 假想起始位置
    id_CS_0 = np.arange(Begin_CS_0, num_C, GirthInterval)  # 假想环缝位置
    print('假想环缝位置',c_un[id_CS_0])
    belong_i = 75  # 强度值搜索半径  # 30
    mid_ = 2  # 余量区间 3
    RB_S = np.array(c_un[0])  # 衬砌开始位置
    RB_E = []  # 衬砌结束位置
    c_name_ins = np.empty(num_CS_0)  # 极低值容器
    c_ = []  # 环缝的存储名
    c_id = []  # 环缝的存储下标器
    N = 8  # 环缝间隔
    #   dis_max = i_c_mean - i_c_std * 3  # 最大差值
    for i in range(num_CS_0):
        # 确保索引在有效范围内
        id_start = max(0, id_CS_0[i] - belong_i)
        id_end = min(len(i_c), id_CS_0[i] + belong_i)
        # 求强度值极低点
        id_min_i_ = np.argmin(i_c[id_start:id_end]) + id_start
        c_name_ins[i] = c_un[id_min_i_]  # 强度值极低点位置
        print('修改后的第', i, '个极低值位置为', c_un[id_min_i_])
        # 寻找以强度值为主的开始和结束位置
        id_min = int(c_name_ins[i] - N)
        id_max = int(c_name_ins[i] + N)
        # 添加衬砌表面起止位置
        RB_E = np.append(RB_E, id_min)  # 添加结束位置
        RB_S = np.append(RB_S, id_max)  # 添加起始位置
    RB_E = np.append(RB_E, c_un[-1])  # 结束位置封顶
    for i in range(num_CS_0):
        if i == 0:
            seams_all = np.arange(RB_E[i], RB_S[i + 1])  # IndexError: too many indices for array: array is 1-dimensional, but 2 were indexed
        else:
            seams_all = np.append(seams_all, np.arange(RB_E[i], RB_S[i + 1]))
    c_xyzic = xyzic[np.isin(xyzic[:, 4], seams_all), :]  # 返回xyzic[:, -1]中有c[c_in]的行数
    # 精简衬砌表面
    id_del = np.isin(np.isin(xyzic[:, 4], seams_all), False)  # 除去环缝点云的点云下标
    txti_delC = xyzic[id_del, :]  # 去除环缝的点云

    return  txti_delC, c_xyzic

def fit_3d_circle(xyz):
    '拟合三维圆'
    num, dim = xyz.shape
    # 求解平面法向量
    L1 = np.ones((num, 1))
    A = np.linalg.inv(xyz.T @ xyz) @ xyz.T @ L1
    # 构建矩阵B和向量L2
    B_rows = (num - 1) * num // 2
    B = np.zeros((B_rows, 3))
    L2 = np.zeros(B_rows)
    count = 0
    for i in range(num):
        for j in range(i + 1, num):
            B[count] = xyz[j] - xyz[i]
            L2[count] = 0.5 * (np.sum(xyz[j] ** 2) - np.sum(xyz[i] ** 2))
            count += 1
    # 构造矩阵D和向量L3
    D = np.zeros((4, 4))
    D[0:3, 0:3] = B.T @ B
    D[0:3, 3] = A.flatten()  # 前三行第四列为A
    D[3, 0:3] = A.T  # 第四行前三列为A的转置
    B_transpose_L2 = B.T @ L2
    L3 = np.concatenate([B_transpose_L2, [1]]).reshape(4, 1)
    # 求解圆心坐标C
    C = np.linalg.inv(D.T) @ L3
    C = C[:3].flatten()  # 提取前三个元素作为圆心
    # 计算半径
    distances = np.linalg.norm(xyz - C, axis=1)
    r = np.mean(distances)
    return np.concatenate([C, [r]])

def fit_3d_circle_mp(xyzc,num_thread=mp.cpu_count()):
    '并行拟合三维圆算法'
    c_un = np.unique(xyzc[:,3])  # 截面序列
    num_c = len(c_un)  # 截面数量
    xyzr_all = np.empty([num_c,4])  # 输出容器
    pool = mp.Pool(processes=num_thread)  # 开启多进程池，数量为cpu
    j = 0  # 分块输出计时器
    # 并行计算
    multi_res = pool.starmap_async(fit_3d_circle, ((xyzc[xyzc[:,3]==c_un[i],:3],) for i in
                 tqdm(range(num_c),desc='分配任务拟合单个三维圆参数',unit='个cross-section',total=num_c)))
    for res in tqdm(multi_res.get(), total=num_c, desc='导出单个三维圆参数', unit='个cross-section'):
        xyzr_all[j,:] = res
        j += 1
    pool.close()  # 禁止进程池再接收任务  #Prohibit process pools from receiving tasks again
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算  #After all processes have completed their calculations, exit parallel computing
    return xyzr_all

def STSD_add_C(las,inx_d = 0.006168):
    '对STSD数据集添加截面序列号'
    # 整理基本信息
    xyz = las.xyz
    I = las.intensity
    inx = las.inx
    label = las.classification
    # 人工制作截面序列号
    inx_un = np.unique(inx)
    inx_min = np.min(inx)
    inx_max = np.max(inx)
    inx_range = np.arange(start=inx_min, stop=inx_max + inx_d, step=inx_d)
    k = 0
    xyzicl = np.c_[xyz,I,inx,label]
    for i in inx_range:
        j = i+inx_d
        xyzicl_ = xyzicl[xyzicl[:,4]<j,:]
        xyzicl_ = xyzicl_[xyzicl_[:, 4] >= i, :]
        xyzicl_[:,4] = k
        if k == 0:
            xyzicl_out = xyzicl_
        else:
            xyzicl_out = np.r_[xyzicl_out,xyzicl_]
        k += 1
    return xyzicl_out

def get_nei_dis_mp(xyz,r,tree,dis_all,cpu=mp.cpu_count()-6):
    '求每个点的平均凸起(建议线程数不超过核心数)'
    # 准备工作
    num = len(xyz)
    dis_in = np.empty(num)  # 存储平均突起的容器
    tik = zs.cut_down(num, cpu)  # 并行计算分块
    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    # 开始进行并行计算
    multi_res = [pool.apply_async(get_nei_dis_block, args=(xyz[tik[i]:tik[i + 1],:], dis_all, tree, r)) for i in
                 range(cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    tik_ = 0  # 计数器
    for res in multi_res:
        dis_in[tik[tik_]:tik[tik_ + 1]] = res.get()
        print('已完成',str(tik[tik_]),'到',str(tik[tik_+1]))
        tik_ += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    return dis_in

def get_nei_dis_block(xyz_,dis_all,tree,r):
    '分块求每个点的平均凸起'
    num_ = len(xyz_)
    dis_in_ = np.zeros(num_)  # 存储平均突起的容器
    for i in tqdm(range(num_)):
        xyz__ = xyz_[i,:]
        indices, dises = tree.query_radius(xyz__.reshape(1, -1), r=r, return_distance=True,
                                           sort_results=True)  # 返回每个点的邻近点下标列表
        indices = indices[0]
        if len(indices) >= 2:
            dis_all_ = dis_all[indices]  # 临近点圆心距  # dis_all是全部的dis_all
            dis_mean = np.mean(dis_all_[1:])  # 当前点平均圆心距
            dis_ = dis_all_[0] - dis_mean  # 求当前点平均突起
            dis_in_[i] = dis_
    return dis_in_

def get_nei_line_density_c(ca_,width):
    '计算截面每个点左右密度差'
    num_ = len(ca_)
    dd_all_ = np.zeros(num_)
    for i in range(num_):
        a_ = ca_[i, 1]
        a_min = a_ - width
        a_max = a_ + width  # 左右角度区间
        count_l = np.sum((ca_[:, 1] > a_min) & (ca_[:, 1] < a_))
        count_r = np.sum((ca_[:, 1] > a_) & (ca_[:, 1] < a_max))  # 左右角度区间点数
        with np.errstate(divide='ignore', invalid='ignore'):
            ratios = np.minimum(count_l, count_r) / np.maximum(count_l, count_r) # 避免除以零的情况
        # 处理NaN和Inf情况（当分母为0时）
        ratios = np.nan_to_num(ratios, nan=0.0, posinf=1.0, neginf=-1.0)
        dd_all_[i] = ratios
    return dd_all_

def get_nei_line_density_mp(ca,width=360/500,cpu=mp.cpu_count()):
    '计算每个点左右密度差(请先对点云按照截面序列号进行排序)'
    # 准备工作
    # num = len(ca)
    # dd_all = np.empty(num)
    # 按照第一列进行排序
    ca = ca[ca[:, 0].argsort()]
    # 统计每个截面的点数
    # unique_sections, section_counts = np.unique(ca[:, 0], return_counts=True)
    unique_sections = np.unique(ca[:, 0])
    num_c = len(unique_sections)
    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    # 开始进行并行计算
    multi_res = pool.starmap_async(get_nei_line_density_c, ((ca[ca[:,0]==unique_sections[i],:], width) for i in
                 tqdm(range(num_c),desc='分配任务给每个截面求左右密度差',unit='cross-sections',total=num_c)))
    j = 0  # 分块输出计时器
    for res in tqdm(multi_res.get(), total=num_c, desc='导出每个点的左右密度差', unit='cross-sections'):
        if j==0:
            dd_all = res
        else:
            dd_all= np.append(dd_all,res)
        j += 1
    pool.close()  # 禁止进程池再接收任务  #Prohibit process pools from receiving tasks again
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算  #After all processes have completed their calculations, exit parallel computing
    return dd_all

def Curvature_r_mp(xyz,r=0.04,cpu=mp.cpu_count()):
    '并行按照球半径计算曲率'
    # 准备工作
    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    num = len(xyz)  # 返回点云数量
    # start, end = block(num, cpu)  # 返回每个block的起始点云下标和终止点云下标
    tik = zr.cut_down(num, cpu)  # 去除bug后的分块函数
    tree = KDTree(xyz)  # 创建树
    j = 0  # 分块输出计数器
    curvature_all = np.empty(shape=len(xyz))  # 新建一个容器：整个点云的曲率数集
    multi_res = [pool.apply_async(curvature_r_block, args=(xyz,tik[i],tik[i+1], tree, r)) for i in range(cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    for res in multi_res:
        curvature_all[tik[j]:tik[j+1]] = res.get()  # 将每个进程得到的结果分block的发送到容器当中
        print('已完成第', tik[j], '至第', tik[j + 1], '的点云')
        j += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    return curvature_all  # 返回全部点云曲率集

def Curvature_r_mp_shm(xyz,r=0.04,cpu=mp.cpu_count()):
    '并行按照球半径计算曲率(共享内存)'
    # 准备工作
    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    num = len(xyz)  # 返回点云数量
    # start, end = block(num, cpu)  # 返回每个block的起始点云下标和终止点云下标
    tik = zr.cut_down(num, cpu)  # 去除bug后的分块函数
    tree = KDTree(xyz)  # 创建树
    j = 0  # 分块输出计数器
    curvature_all = np.empty(shape=len(xyz))  # 新建一个容器：整个点云的曲率数集
    '创建共享内存'
    shm = shared_memory.SharedMemory(create=True, size=xyz.nbytes)
    #  将数据复制到共享内存
    shared_array = np.ndarray(xyz.shape, dtype=xyz.dtype, buffer=shm.buf)
    shared_array[:] = xyz[:]
    print(f"共享内存创建完成: {shm.name}")
    print(f"共享内存大小: {shm.size} 字节")
    multi_res = [pool.apply_async(curvature_r_block, args=(shared_array,tik[i],tik[i+1], tree, r)) for i in range(cpu)]  # 将每个block需要处理的点云区间发送到每个进程当中
    for res in multi_res:
        curvature_all[tik[j]:tik[j+1]] = res.get()  # 将每个进程得到的结果分block的发送到容器当中
        print('已完成第', tik[j], '至第', tik[j + 1], '的点云')
        j += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    # 清理共享内存
    shm.close()
    shm.unlink()  # 重要：释放共享内存
    return curvature_all  # 返回全部点云曲率集

def curvature_r_block(xyz,start,end,tree,r):
    '分块按照球半径计算曲率'
    # xyz_32 = np.column_stack((xyz[:, 0], xyz[:, 1], xyz[:, 2])).astype(np.float32)
    # pcd = o3d.geometry.PointCloud()
    # pcd.points = o3d.utility.Vector3dVector(xyz_32)
    # pcd_tree = o3d.geometry.KDTreeFlann(pcd)
    curvature_all = np.empty(end-start)
    # num_ = len(xyz_)
    # curvature_all = np.empty(num_)
    j = 0
    for i in tqdm(range(start,end)):
        # [_, idx, _] = pcd_tree.search_knn_vector_3d(pcd.points[i], n)  # 求每个点最近的n个点
        # id_kntree[i, :] = idx  # 求出每个点最邻近的n个点的下标
        xyz_ = xyz[i,:]
        indices, dises = tree.query_radius(xyz_.reshape(1, -1), r=r, return_distance=True,
                                           sort_results=True)  # 返回每个点的邻近点下标列表
        indices = indices[0]
        xyz_n = xyz[indices, :]
        cv, _ = zr.pca(xyz_n)  # 求每个点的特征值和特征向量
        c = zr.curvature_(cv,-1)  # 求出当前点的曲率
        # print(c)
        curvature_all[j] = c
        j += 1
    return curvature_all


def Curvature_r(xyz,r=0.04,job=-1):
    '按照球半径计算曲率'
    num = len(xyz)
    curvature_all = np.empty(num)
    tree = KDTree(xyz)  # 创建树
    for i in tqdm(range(num)):
        xyz_ = xyz[i,:]
        indices, dises = tree.query_radius(xyz_.reshape(1, -1), r=r, return_distance=True,
                                           sort_results=True)  # 返回每个点的邻近点下标列表
        indices = indices[0]
        xyz_n = xyz[indices, :]
        cv, _ = zr.pca(xyz_n)  # 求每个点的特征值和特征向量
        c = zr.curvature_(cv,job)  # 求出当前点的曲率
        # print(c)
        curvature_all[i] = c

    return curvature_all

def get_JCFI_mp(xyzic: Union[list, np.ndarray],ps_out: Union[list, np.ndarray], r: float,cpu:int = mp.cpu_count()) ->np.ndarray:
    '计算JCFI指数'
    '1.求曲率'
    C_all = Curvature_r_mp(xyzic[:, :3], r=0.04,cpu=cpu)
    '2.求平均凸起'
    ps = np.r_[xyzic,ps_out]
    del ps_out
    xzrc = zs.fit_circle(xyzic[:, 0], xyzic[:, 2], xyzic[:, 4], num_cpu=cpu)
    # xzry = zs.fit_circle(xyzic[:, 0], xyzic[:, 2], xyzic[:, 1], num_cpu=cpu)
    # xyziy = np.c_[ps[:, :4], ps[:, 1]]
    # dis_all = zs.get_CenterDis(xzry,xyziy,cpu_count=cpu)
    dis_all = zs.get_CenterDis(xzrc, ps, cpu_count=cpu)
    # tree = KDTree(xyzic)
    # del xzry,xyziy
    tree = KDTree(ps[:, :3])  # 创建树
    del ps
    dis_in = get_nei_dis_mp(xyzic[:, :3], r, tree, dis_all, cpu=cpu)
    '3.求点云左右密度差'
    # 求点云角度
    x0 = np.mean(xzrc[:, 0])
    z0 = np.mean(xzrc[:, 1])
    angle_all = zs.get_angle_all(xyzic[:,[0,2]],x0,z0,cpu_count=cpu)
    ps_Cda = np.c_[xyzic,C_all,dis_in,angle_all]
    ca = np.c_[xyzic[:,4],angle_all]
    dd_all = get_nei_line_density_mp(ca,cpu=cpu)
    ps_Cda = ps_Cda[ps_Cda[:, 4].argsort()]
    xyzicJ = np.c_[ps_Cda[:,:7],dd_all,ps_Cda[:,5]*ps_Cda[:,6]*dd_all]
    # xyzicJ = np.c_[xyzic,ps_Cda[:,5]*ps_Cda[:,6]*dd_all]
    # 将 NaN 替换为 0
    xyzicJ = np.nan_to_num(xyzicJ, nan=0.0)
    return xyzicJ

def get_JCFI_noout(xyzic: Union[list, np.ndarray], r: float, cpu:int = mp.cpu_count(), R=2.7) ->np.ndarray:
    '求JCFI(无非衬砌点情况)'
    '1.求曲率'
    C_all = Curvature_r_mp_shm(xyzic[:, :3], r=r,cpu=cpu)
    '2.求平均凸起'
    xzrc = zs.fit_circle(xyzic[:, 0], xyzic[:, 2], xyzic[:, 4], num_cpu=cpu)
    dis_all = zs.get_CenterDis(xzrc, xyzic, cpu_count=cpu)
    tree = KDTree(xyzic[:, :3])  # 创建树
    dis_in = get_nei_dis_mp(xyzic[:, :3], r, tree, dis_all, cpu=cpu)
    '3.求点云左右密度差'
    x0 = np.mean(xzrc[:, 0])
    z0 = np.mean(xzrc[:, 1])
    angle_all = zs.get_angle_all(xyzic[:,[0,2]],x0,z0,cpu_count=cpu)
    ps_Cda = np.c_[xyzic,C_all,dis_in,angle_all]
    ca = np.c_[xyzic[:,4],angle_all]
    dd_all = get_nei_line_density_mp(ca,cpu=cpu)
    ps_Cda = ps_Cda[ps_Cda[:, 4].argsort()]
    '4.计算周长'
    # perimeter = angle_all / 360 * 2 * np.pi * R  # 基于周长求各点实际位置
    # perimeter = perimeter[ps_Cda[:, 4].argsort()]
    angle_all = zs.get_angle_all(ps_Cda[:, [0, 2]], x0, z0, cpu_count=cpu)
    perimeter = angle_all / 360 * 2 * np.pi * R
    # xyzicJ = np.c_[xyzic, ps_Cda[:, 5] * ps_Cda[:, 6] * dd_all]
    # xyzicJ = np.c_[ps_Cda[:, :7], dd_all, ps_Cda[:, 5] * ps_Cda[:, 6] * dd_all]
    xyzicJP = np.c_[ps_Cda[:,:5], ps_Cda[:, 5] * ps_Cda[:, 6] * dd_all,perimeter]
    # 将 NaN 替换为 0
    xyzicJP = np.nan_to_num(xyzicJP, nan=0.0)
    return xyzicJP

def get_JCFI_ZF(xyzic: Union[list, np.ndarray], r: float,cpu:int = mp.cpu_count(),R=2.7) ->np.ndarray:
    '求JCFI(纵缝版)'
    '1.求曲率'
    C_all = Curvature_r_mp_shm(xyzic[:, :3], r=r,cpu=cpu)
    '2.求平均凸起'
    xzrc = zs.fit_circle(xyzic[:, 0], xyzic[:, 2], xyzic[:, 4], num_cpu=cpu)
    dis_all = zs.get_CenterDis(xzrc, xyzic, cpu_count=cpu)
    tree = KDTree(xyzic[:, :3])  # 创建树
    dis_in = get_nei_dis_mp(xyzic[:, :3], r, tree, dis_all, cpu=cpu)
    '3.求点云左右密度差'
    x0 = np.mean(xzrc[:, 0])
    z0 = np.mean(xzrc[:, 1])
    angle_all = zs.get_angle_all(xyzic[:,[0,2]],x0,z0,cpu_count=cpu)
    ps_Cda = np.c_[xyzic,C_all,dis_in,angle_all]
    ca = np.c_[xyzic[:,4],angle_all]
    dd_all = get_nei_line_density_mp(ca,cpu=cpu)
    ps_Cda = ps_Cda[ps_Cda[:, 4].argsort()]
    # xyzicJ = np.c_[xyzic, ps_Cda[:, 5] * ps_Cda[:, 6] * dd_all]
    # xyzicJ = np.c_[ps_Cda[:, :7], dd_all, ps_Cda[:, 5] * ps_Cda[:, 6] * dd_all]
    '4.计算周长'
    # perimeter = angle_all / 360 * 2 * np.pi * R  # 基于周长求各点实际位置
    # perimeter = perimeter[ps_Cda[:, 4].argsort()]
    angle_all = zs.get_angle_all(ps_Cda[:, [0, 2]], x0, z0, cpu_count=cpu)
    perimeter = angle_all / 360 * 2 * np.pi * R
    xyzicJpC = np.c_[ps_Cda[:,:5], ps_Cda[:, 5] * ps_Cda[:, 6] * dd_all, perimeter,ps_Cda[:, 5]]
    # 将 NaN 替换为 0
    xyzicJpC = np.nan_to_num(xyzicJpC, nan=0.0)
    return xyzicJpC

def get_VY(shared_array,r,tree,i):
    '计算单点的Y方向分量'
    indices= tree.query_radius(shared_array[i,:3].reshape(1, -1), r=r, return_distance=False,
                                       sort_results=False)  # 返回每个点的邻近点下标列表
    indices = indices[0]
    xyz_ = shared_array[indices,:3]
    if len(xyz_)>=3:
        cov = np.cov(xyz_.T)
        eigenvals, eigenvecs = np.linalg.eigh(cov)
        v1 = eigenvecs[:, -1]  # 最大特征值对应的方向（主方向）
        vY = v1[1]
    else:
        vY = 0
    return vY

def get_HF_bulge(c_all,c_interval=500,c_long=20,num_min=40):
    '基于环缝凸起提取'
    # 整理截面序列与点数量
    c_un, num_c = np.unique(c_all, return_counts=True)
    # print(np.mean(num_c))
    # print(np.std(num_c))
    cn = np.c_[c_un,num_c]
    # 找到最大索引
    ind_max = np.argmax(num_c)
    c_max = c_un[ind_max]
    c_max_l = c_max-c_long
    c_max_r = c_max+c_long  # 假定左右缝极限
    # 筛选范围内的数据
    mask = (c_un >= c_max_l) & (c_un <= c_max_r) & (num_c >= num_min)
    # 获取满足条件的c_un值
    c_un_min = np.min(c_un)
    c_un_max = np.max(c_un)
    c_num_long = int(np.ceil((c_un_max-c_un_min)/c_interval))  # 假想环缝数量
    c_lf_all = np.zeros([c_num_long,2])  # 环缝起始和结束位置容器
    # valid_c_un = c_un[mask]  # 极低值起始和结束
    ind_c0_begin = (c_max-c_un_min) % c_interval + c_un_min # 假想起始位置
    for i in range(c_num_long):
        id_min_i_ =  (c_un >= (ind_c0_begin-c_long*7)) & (c_un <= (ind_c0_begin+c_long*7))  # 搜索极低值半径
        cn_ = cn[id_min_i_,:]  # 搜索半径内的数据
        id_max_ = np.argmax(cn_[:,1])  # 点数量最多的位置
        c_name_max = cn_[id_max_,0] # 所在截面序列号
        mask_ = (c_un >= c_name_max-c_long) & (c_un <= c_name_max+c_long) & (num_c >= num_min)  # 复合条件的索引
        valid_c_un_ = c_un[mask_]  # 极低值起始和结束
        c_lf_all[i,0]=valid_c_un_[0]
        c_lf_all[i,1]=valid_c_un_[-1]  # 环缝起始和结束位置
        # 刷新假想位置
        ind_c0_begin += c_interval
    # 创建所有范围的条件
    # 方法1：删除起始值和结束值都为0的行
    mask = ~((c_lf_all[:, 0] == 0) & (c_lf_all[:, 1] == 0))
    c_lf_all = c_lf_all[mask]
    print('环缝截面区间',c_lf_all)
    conditions = [(c_all >= start) & (c_all <= end)
                  for start, end in c_lf_all]
    # 合并所有条件
    final_mask = np.logical_or.reduce(conditions)
    inverse_mask = ~final_mask
    return final_mask, inverse_mask, c_lf_all

def get_HF_bulge_scipy(c_all,c_long=20,num_min=40):
    '基于环缝凸起提取'
    c_un, num_c = np.unique(c_all, return_counts=True)
    peaks, properties = find_peaks(num_c,height=np.mean(num_c) * 0.5,  # 最小峰值高度
                                   distance=int(len(num_c) * 0.1),  # 峰值间最小距离
                                   prominence=np.std(num_c) * 0.3)  # 峰值突出度
    c_un_maxes = c_un[peaks]  # 峰值对应的c值
    num_c_un_maxes = len(c_un_maxes)  # 峰值的数量
    c_lf_all = np.zeros([num_c_un_maxes, 2])  # 环缝起始和结束位置容器
    for i in range(num_c_un_maxes):
        mask_ = (c_un >= c_un_maxes[i]-c_long) & (c_un <= c_un_maxes[i]+c_long) & (num_c >= num_min)  # 复合条件的索引
        valid_c_un_ = c_un[mask_]  # 极低值起始和结束
        c_lf_all[i,0]=valid_c_un_[0]
        c_lf_all[i,1]=valid_c_un_[-1]  # 环缝起始和结束位置
    print('环缝截面区间', c_lf_all)
    conditions = [(c_all >= start) & (c_all <= end)
                  for start, end in c_lf_all]
    # 合并所有条件
    final_mask = np.logical_or.reduce(conditions)
    inverse_mask = ~final_mask
    return final_mask, inverse_mask, c_lf_all

def get_HF_JCFI(ps,r=0.07,cpu=mp.cpu_count()):
    '基于JCFI提取环缝'
    # 1.将大于阈值的点提取出来
    mean_J = np.mean(ps[:,5])  # JCFI均值
    std_J = np.std(ps[:,5])  # JCFI标准差
    v_J = mean_J+std_J*3
    # v_J = 0.079
    print('均值为',mean_J,'标准差为',std_J,'阈值为',v_J)
    # 剔除第一波点云
    ps_J = ps[ps[:, 5] > v_J, :]  # 符合条件的环缝
    ps_no_1 = ps[ps[:,5]<=v_J,:] # 第一轮被剔除的环缝
    # 2.计算ps_J在Y轴方向的分量
    tree = KDTree(ps_J[:, :3])  # 创建树
    num_J = len(ps_J)  # 种子点数量
    vectors_Y = np.zeros(num_J)  # Y方向分量
    '创建共享内存'
    shm = shared_memory.SharedMemory(create=True, size=ps_J.nbytes)
    #  将数据复制到共享内存
    shared_array = np.ndarray(ps_J.shape, dtype=ps_J.dtype, buffer=shm.buf)
    shared_array[:] = ps_J[:]
    print(f"共享内存创建完成: {shm.name}")
    print(f"共享内存大小: {shm.size} 字节")
    '开启并行计算'
    pool = mp.Pool(processes=cpu)
    multi_res = pool.starmap_async(get_VY, ((shared_array,r,tree,i) for i in
                 tqdm(range(num_J),desc='分配任务计算单点Y方向分量',unit='个点',total=num_J)))
    j = 0
    for res in tqdm(multi_res.get(),total=num_J,desc='输出Y方向分量'):
        vectors_Y[j] = res
        j += 1
    pool.close()  # 关闭所有进程
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    # 清理共享内存
    shm.close()
    shm.unlink()  # 重要：释放共享内存
    # 3.高斯混合模型分解成三类，要接近0那类
    labels = zm.use_gmm(vectors_Y.reshape(-1, 1),3)
    gmm_0 = vectors_Y[labels == 0]
    gmm_1 = vectors_Y[labels == 1]
    gmm_2 = vectors_Y[labels == 2]
    # 计算每个类别的平均值
    mean_0 = np.mean(gmm_0)
    mean_1 = np.mean(gmm_1)
    mean_2 = np.mean(gmm_2)
    # 计算与0的距离（绝对值）
    dist_0 = abs(mean_0)
    dist_1 = abs(mean_1)
    dist_2 = abs(mean_2)
    # 找到距离0最近的类别
    distances = [dist_0, dist_1, dist_2]
    min_index = distances.index(min(distances))
    # 根据索引选择对应的数组
    if min_index == 0:
        ps_gmm = ps_J[labels == 0, :]
        ps_no_2 = ps_J[labels == 1, :]
        ps_no_3 = ps_J[labels == 2, :]
    elif min_index == 1:
        ps_gmm = ps_J[labels == 1, :]
        ps_no_2 = ps_J[labels == 0, :]
        ps_no_3 = ps_J[labels == 2, :]
    else:
        ps_gmm = ps_J[labels == 2, :]  # 为下一步准备的点云
        ps_no_2 = ps_J[labels == 0, :]
        ps_no_3 = ps_J[labels == 1, :]  # 剔除的点云
    # 4. 实例分割并返回环缝截面序列
    in_index,out_index,hf_lf = get_HF_bulge(ps_gmm[:,4])
    # 5.输出环缝点云，非环缝点云，和环缝点云截面区间集合
    ps_hf = ps_gmm[in_index,:]
    ps_no_4 = ps_gmm[out_index,:]
    ps_no = np.r_[ps_no_1,ps_no_2,ps_no_3,ps_no_4]
    return ps_hf,ps_no,hf_lf

def cut_DGH(xyzic,hf_lf):
    '分割盾构环'
    num_DGH = len(hf_lf)
    list_DGH = []
    for i in range(num_DGH):
        xyzic_l = xyzic[xyzic[:,4]<hf_lf[i,0],:]
        list_DGH.append(xyzic_l)
        xyzic=xyzic[xyzic[:,4]>=hf_lf[i,1],:]
    list_DGH.append(xyzic)
    return list_DGH

# def fit_line_by2ps(oy):
#     '两个最远点计算直线kb'
#     ind_y_max = np.argmax(oy[:,1])
#     ind_y_min = np.argmin(oy[:,1])
#     oy_max = oy[ind_y_max,:]
#     oy_min = oy[ind_y_min,:]
#     k = (oy_max[1]-oy_min[1])/(oy_max[0]-oy_min[0])
#     print(k)

def Merge_straight_lines(lines, threshold = 10):
    '合并相同直线'
    distances = squareform(pdist(lines))  # 距离矩阵
    groups = []
    visited = set()
    for i in range(len(lines)):
        if i in visited:
            continue
        group = [i]
        visited.add(i)
        for j in range(i + 1, len(lines)):
            if j not in visited and distances[i, j] < threshold:
                group.append(j)
                visited.add(j)
        groups.append(group)
    print("按距离分组结果：")
    for idx, group in enumerate(groups):
        print(f"组 {idx}: {group}")
    return groups

def get_ZF_JCFI(xyzic,cpu=mp.cpu_count(),r=0.03,eps=0.02,min_samples=4,sigma=0.01):
    '基于JCFI提取每个盾构环的纵缝'
    #1. 计算JCFI均值与标准差
    xyzicjp = get_JCFI_noout(xyzic, r,cpu)
    mean_JCFI = np.mean(xyzicjp[:,5])
    std_JCFI = np.std(xyzicjp[:,5])
    td0 = mean_JCFI+3*std_JCFI  # 严格阈值
    print('严格JCFI阈值',td0)
    #3. 通过DBSCAN聚类并剔除非种子点
    xyzicj_td0 = xyzicjp[xyzicjp[:,5]>=td0,:]
    clustering = DBSCAN(eps=eps, min_samples=min_samples).fit(xyzicj_td0[:,:3])
    labels = clustering.labels_  # 每个点的类别标签，-1 表示噪声
    xyzicjzl = np.c_[xyzicj_td0,labels]
    xyzicjzl = xyzicjzl[xyzicjzl[:,-1]>=0,:]
    l_un,num_l = np.unique(xyzicjzl[:,-1],return_counts=True)
    seed_index = []  #  种子索引
    C_len = len(np.unique(xyzicjzl[:, 4]))
    for i in l_un:
        xyzicjzl_ = xyzicjzl[xyzicjzl[:,-1]==i,:]
        num_c_ = len(np.unique(xyzicjzl_[:,4]))
        if num_c_>=0.25*C_len:
            seed_index.append(i)
    print('候选种子点索引',seed_index)
    seed_index_in = np.isin(xyzicjzl[:, -1], seed_index)
    xyzicjzl_in = xyzicjzl[seed_index_in, :]
    # zt.view_pointclouds(xyzicjzl_in[:, :3], xyzicjzl_in[:, -1], colormap='hsv')
    print('候选种子点数量',xyzicjzl_in.shape)
    num_seed_index = len(seed_index)
    #4. 对各个联通区域拟合直线pY
    kb_all = np.empty([num_seed_index, 2])  # 直线属性容器
    j = 0
    for i in seed_index:
        xyzicjzl_in_ = xyzicjzl_in[xyzicjzl_in[:,-1]==i,:]
        oy_ = xyzicjzl_in_[:,[6,1]]
        # 二维直线拟合
        kb_all[j,:] = zR.fit_2Dline_ransac(oy_,sigma=sigma)
        j+=1
    #5. 合并相同直线
    groups_lines = Merge_straight_lines(kb_all, 2)
    #6. 重新整理种子点数据
    kb_list = np.empty([len(groups_lines), 2])
    for idx, group in enumerate(groups_lines):
        if len(group)>=2:
            selected_labels = np.array(seed_index)[group]
            xyzicjzl_in_ = xyzicjzl_in[np.isin(xyzicjzl_in[:,-1],selected_labels),:]
            oy_ = xyzicjzl_in[:, [6, 1]]
            kb_list[idx,:] = zR.fit_2Dline_ransac(oy_,sigma=sigma)
        else:
            kb_list[idx, :] = kb_all[idx, :]
    print('联通区域拟合直线属性',kb_list)
    # #7. 找到直线区域平均距离大于0的点数量
    # # 点到直线的距离
    # zf_list =[]
    # td1 = mean_JCFI+1*std_JCFI  # 不严格阈值
    # oy_all = xyzicjz[:,[6,1]]
    # td2 = 0.1  # 距离阈值
    # for i in range(len(kb_list)):
    #     dis_all_ = zm.get_distance_point2line(oy_all,kb_list[i])
    #     ps_ = xyzicjz[dis_all_<td2,:]
    #     ps_ = ps_[ps_[:,5]>td1,:]
    #     num_C_ = len(np.unique(ps_[:,4]))  # 截面数量
    #     if num_C_/C_len >0.5:
    #         zf_list.append(ps_)
    # zf_array = np.vstack(zf_list)
    # 找到符合条件的下标并合并
    zf_bool = np.zeros(len(xyzicjp), dtype=bool)
    td1 = mean_JCFI + 1 * std_JCFI  # 不严格阈值
    td_dis = 0.1  # 距离阈值
    print('不严格JCFI阈值',td1)
    py_all = xyzicjp[:,[6,1]]
    JCFI_all = xyzicjp[:,5]
    for i in range(len(kb_list)):
        dis_all_ = zm.get_distance_point2line(py_all, kb_list[i])
        # 限制条件1
        valid_indices = dis_all_ < td_dis
        # 限制条件2
        condition_mask = JCFI_all > td1
        # 同时满足两个条件的位置
        final_mask = valid_indices & condition_mask
        ps_ = xyzicjp[final_mask,:]
        num_C_ = len(np.unique(ps_[:, 4]))  # 截面数量
        if num_C_ / C_len > 0.5:
            # 将满足条件的位置在zf_bool中设为True
            zf_bool = zf_bool | final_mask  # 使用或操作，累积所有满足条件的位置
    # 返回zf_bool为True的点云
    xyzic_true = xyzicjp[zf_bool, :]
    # 返回zf_bool为False的点云
    xyzic_false = xyzicjp[~zf_bool, :]
    return xyzic_true, xyzic_false

def get_ZF_C(xyzic,cpu=mp.cpu_count(),r=0.035,eps=0.02,min_samples=4,sigma=0.01,R=2.7,td_dis=0.1):
    '基于球曲率求单个盾构环的纵缝'
    '1.计算JCFI和周长'
    xyzicJpC = get_JCFI_ZF(xyzic,r=r,cpu=cpu,R=R)
    # 计算JCFI的严格阈值
    mean_JCFI = np.mean(xyzicJpC[:,5])
    std_JCFI = np.std(xyzicJpC[:,5])
    td0 = mean_JCFI+3*std_JCFI  # 严格阈值
    print('严格JCFI阈值',td0)
    # 寻找第一批候选点
    # xyzicp = np.c_[xyzic,perimeter]
    ps_td0 = xyzicJpC[xyzicJpC[:,5] >= td0, :]
    '2.DBSCAN聚类'
    clustering = DBSCAN(eps=eps, min_samples=min_samples).fit(ps_td0[:,:3])
    labels = clustering.labels_  # 每个点的类别标签，-1 表示噪
    ps_l = np.c_[ps_td0,labels]
    ps_l = ps_l[ps_l[:, -1] >= 0, :]
    l_un,num_l = np.unique(ps_l[:,-1],return_counts=True)
    seed_index = []  #  种子索引
    C_len = len(np.unique(xyzic[:, 4]))  # 盾构环截面数量
    for i in l_un:
        ps_l_ = ps_l[ps_l[:,-1]==i,:]
        num_c_ = len(np.unique(ps_l_[:,4]))
        if num_c_>=0.25*C_len:
            seed_index.append(i)
    print('候选种子联通索引',seed_index)
    seed_index_in = np.isin(ps_l[:, -1], seed_index)
    ps_in = ps_l[seed_index_in, :]
    print('候选种子点数量', ps_in.shape)
    num_seed_index = len(seed_index)  # 联通区域数量
    '3.对各个联通区域拟合直线pY，合并相近直线'
    # 拟合直线
    kb_all = np.empty([num_seed_index, 2])  # 直线属性容器
    j = 0
    for i in seed_index:
        ps_in_ = ps_in[ps_in[:,-1]==i,:]
        py_ = ps_in_[:,[6,1]]
        # 二维直线拟合
        kb_all[j,:] = zR.fit_2Dline_ransac(py_,sigma=sigma)
        j+=1
    print(kb_all)
    # 合并相同直线
    groups_lines = Merge_straight_lines(kb_all, 2)
    '4.整理最后的种子点数据'
    # 重新整理种子点数据
    kb_list = np.empty([len(groups_lines), 2])
    for idx, group in enumerate(groups_lines):
        if len(group)>=2:
            selected_labels = np.array(seed_index)[group]
            ps_in_ = ps_in[np.isin(ps_in[:,-1],selected_labels),:]
            py_ = ps_in_[:, [6, 1]]
            kb_list[idx,:] = zR.fit_2Dline_ransac(py_,sigma=sigma)
        else:
            kb_list[idx, :] = kb_all[idx, :]
    print('联通区域拟合直线属性',kb_list)
    '5.输出纵缝点和衬砌点'
    # 找到符合条件的下标并合并
    zf_bool = np.zeros(len(xyzic), dtype=bool)
    mean_C = np.mean(xyzicJpC[:,7])
    std_C = np.std(xyzicJpC[:,7])
    td1 = mean_C+3*std_C  # 不严格阈值
    print('严格球曲率阈值',td1)
    py_all = xyzicJpC[:,[6,1]]
    C_all = xyzicJpC[:,7]
    for i in range(len(kb_list)):
        dis_all_ = zm.get_distance_point2line(py_all, kb_list[i])
        # 限制条件1
        valid_indices = dis_all_ < td_dis
        # 限制条件2
        condition_mask = C_all > td1
        # 同时满足两个条件的位置
        final_mask = valid_indices & condition_mask
        ps_ = xyzicJpC[final_mask,:]
        num_C_ = len(np.unique(ps_[:, 4]))  # 截面数量
        if num_C_ / C_len > 0.5:
            # 将满足条件的位置在zf_bool中设为True
            zf_bool = zf_bool | final_mask  # 使用或操作，累积所有满足条件的位置
        # np.savetxt('E:\\2025博二上学期\\基于复合指数的RMLS盾构隧道环缝和纵缝提取\\Data\\test_ps_.txt',ps_,fmt='%.05f')
    # 返回zf_bool为True的点云
    xyzic_true = xyzicJpC[zf_bool, :]
    # 返回zf_bool为False的点云
    xyzic_false = xyzicJpC[~zf_bool, :]
    # np.savetxt('E:\\2025博二上学期\\基于复合指数的RMLS盾构隧道环缝和纵缝提取\\Data\\xyzic_true.txt',xyzic_true,fmt='%.05f')
    # np.savetxt('E:\\2025博二上学期\\基于复合指数的RMLS盾构隧道环缝和纵缝提取\\Data\\xyzic_false.txt',xyzic_false,fmt='%.05f')
    return xyzic_true, xyzic_false

def get_HF_intensity_peaks(ps,length_05 = 15,dis=200):
    '251103强化版本强度值环缝提取算法'
    c_un = np.unique(ps[:,4])
    num_C = len(c_un)
    # 计算每个环的平均强度值
    i_c = np.empty(num_C)
    # 并行计算准备
    tik = zs.cut_down(num_C)  # 分块起止点
    pool = mp.Pool(processes=mp.cpu_count())  # 开启多进程池，数量为cpu
    multi_res = [pool.apply_async(zs.find_cImean_block, args=(ps, c_un, tik[i], tik[i + 1])) for i in  # points_new
                 range(mp.cpu_count())]  # 将每个block需要处理的点云区间发送到每个进程当中
    tik_ = 0  # 计数器
    for res in multi_res:
        i_c[tik[tik_]:tik[tik_ + 1]] = res.get()
        tik_ += 1
    pool.close()  # 禁止进程池再接收任务
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    '归一化特征值'
    i_c = zm.normalization(i_c,1.0)
    # 取反
    i_c_1 = 1-i_c
    '计算极低值'
    peaks, properties = find_peaks(i_c_1,height=np.mean(i_c_1)*0.5,  # 最小峰值高度
                                   distance=dis,  # 峰值间最小距离
                                   prominence=np.std(i_c_1) * 0.3)  # 峰值突出度
    c_un_maxes = c_un[peaks]  # 峰值对应的c值
    print('环缝中心截面位置',c_un_maxes)
    num_hf = len(c_un_maxes)  # 环缝数量
    # length_05 = 15  # 假定环缝半宽度
    td = 3
    hf_lf = np.empty([num_hf,2])
    for i in range(num_hf):
        # 找到左右假定搜索边界
        peaks_l_ = peaks[i]-length_05
        peaks_r_ = peaks[i]+length_05
        # 找到左右非环缝边界
        cq_l_ = peaks_l_-length_05
        cq_r_ = peaks_r_+length_05
        # --- 新增：边界检查与修正 ---
        peaks_l_ = max(0, peaks_l_)
        peaks_r_ = min(num_C-1, peaks_r_)  # 索引最大为 num_C - 1
        cq_l_ = max(0, cq_l_)
        cq_r_ = min(num_C-1, cq_r_)  # 索引最大为 num_C - 1
        # --- 结束新增 ---
        # 求左右衬砌的平均标准差和阈值
        i_c_l = i_c[cq_l_:peaks_l_]
        i_c_r = i_c[peaks_r_:cq_r_]
        mean_i_c_l = np.mean(i_c_l)
        std_i_c_l = np.std(i_c_l)
        mean_i_c_r = np.mean(i_c_r)
        std_i_c_r = np.std(i_c_r)
        td_l_ = mean_i_c_l - std_i_c_l * td
        td_r_ = mean_i_c_r - std_i_c_r * td
        # 左右待选值
        i_c_hf_l = i_c[peaks_l_:peaks[i]]
        i_c_hf_r = i_c[peaks[i]:peaks_r_]
        # 在 i_c_l 中查找最后一个小于 td_l_ 的元素的相对下标
        indices_l = np.where(i_c_hf_l < td_l_)[0]
        if len(indices_l) > 0:
            last_idx_l_rel = indices_l[0]
        else:
            last_idx_l_rel = 0
        indices_r = np.where(i_c_hf_r < td_r_)[0]
        if len(indices_r) > 0:
            first_idx_r_rel = indices_r[-1]
        else:
            first_idx_r_rel = 14
        # 整理左右下标
        idx_l_ = peaks[i] - length_05 + last_idx_l_rel
        idx_r_ = peaks[i] + first_idx_r_rel
        idx_r_ = min(idx_r_, num_C-1)
        hf_lf[i,0] = c_un[idx_l_]
        hf_lf[i,1] = c_un[idx_r_]
    print('环缝截面起止位置',hf_lf)
    # 确保 hf_lf 是整数类型，因为它们将用作索引
    hf_lf = hf_lf.astype(int)
    # 假设 ps[:, 4] 存储的是环的索引 c
    c_indices = ps[:, 4]
    # 创建一个布尔掩码，标记哪些点属于环缝
    mask_hf = np.zeros(len(ps), dtype=bool)
    for start, end in hf_lf:
        # 创建当前环缝范围的掩码，并与总掩码进行或运算
        mask_hf |= (c_indices >= start) & (c_indices <= end)
    # 使用掩码分割数据
    ps_hf = ps[mask_hf]  # 环缝部分
    ps_nf = ps[~mask_hf]  # 非环缝部分 (~ 是逻辑非运算符)
    return ps_hf, ps_nf




import numpy as np  # 数组库
import multiprocessing as mp  # 并行计算库
import math  # 添加数学库为cuda编程做准备
from tqdm import tqdm  # 进度条

class supervoxel:
    '建立超体素'

    def __init__(self, xyz):
        '构造函数'
        self.xyz = xyz


from __future__ import division
# 这是一个关于私人用途的代码原文件库
import numpy as np
from numpy import *
import sys  # maxint
import matplotlib.pyplot as plt
import pyvista as pv
import polars as pl
from PIL import Image
import plyfile as pf
import os
from scipy.interpolate import griddata
import math

# C-均值聚类
def FCM_train(X, n_centers, m, max_iter=100, theta=1e-5, seed=0):
    rng = np.random.RandomState(seed)
    N, D = np.shape(X)
    # 随机初始化关系矩阵
    U = rng.uniform(size=(N, n_centers))
    # 保证每行和为1
    U = U / np.sum(U, axis=1, keepdims=True)
    # 开始迭代
    for i in range(max_iter):
        print(i)
        U_old = U.copy()
        centers = FCM_getCenters(U, X, m)
        U = FCM_getU(X, centers, m)
        # 两次关系的距离矩阵过小，结束训练
        if np.linalg.norm(U - U_old) < theta:
            break
    return centers, U


def FCM_getCenters(U, X, m):
    N, D = np.shape(X)
    N, C = np.shape(U)
    um = U ** m
    tile_X = np.tile(np.expand_dims(X, 1), [1, C, 1])
    tile_um = np.tile(np.expand_dims(um, 1), [1, 1, D])
    temp = tile_X * tile_um
    new_C = np.sum(temp, axis=0) / np.expand_dims(np.sum(um, axis=0), axis=-1)
    return new_C


def FCM_getU(X, Centers, m):
    N, D = np.shape(X)
    C, D = np.shape(Centers)
    temp = FCM_dist(X, Centers) ** float(2 / (m - 1))
    tile_temp = np.tile(np.expand_dims(temp, 1), [1, C, 1])
    denominator_ = np.expand_dims(temp, -1) / tile_temp
    return 1 / np.sum(denominator_, axis=-1)

def FCM_dist(X, Centers):
    N, D = np.shape(X)
    C, D = np.shape(Centers)
    tile_x = np.tile(np.expand_dims(X, 1), [1, C, 1])
    tile_centers = np.tile(np.expand_dims(Centers, axis=0), [N, 1, 1])
    dist = np.sum((tile_x - tile_centers) ** 2, axis=-1)
    return np.sqrt(dist)

def qhull2D(sample):
    link = lambda a, b: concatenate((a, b[1:]))
    edge = lambda a, b: concatenate(([a], [b]))

    def dome(sample, base):
        h, t = base
        dists = dot(sample - h, dot(((0, -1), (1, 0)), (t - h)))
        outer = repeat(sample, dists > 0, 0)
        if len(outer):
            pivot = sample[argmax(dists)]
            return link(dome(outer, edge(h, pivot)),
                        dome(outer, edge(pivot, t)))
        else:
            return base

    if len(sample) > 2:
        axis = sample[:, 0]
        base = take(sample, [argmin(axis), argmax(axis)], 0)
        return link(dome(sample, base), dome(sample, base[::-1]))
    else:
        return sample


def minBoundingRect(hull_points_2d):
    # print "Input convex hull points: "
    # print hull_points_2d

    # Compute edges (x2-x1,y2-y1)
    edges = zeros((len(hull_points_2d) - 1, 2))  # empty 2 column array
    for i in range(len(edges)):
        edge_x = hull_points_2d[i + 1, 0] - hull_points_2d[i, 0]
        edge_y = hull_points_2d[i + 1, 1] - hull_points_2d[i, 1]
        edges[i] = [edge_x, edge_y]
    # print "Edges: \n", edges

    # Calculate edge angles   atan2(y/x)
    edge_angles = zeros((len(edges)))  # empty 1 column array
    for i in range(len(edge_angles)):
        edge_angles[i] = math.atan2(edges[i, 1], edges[i, 0])
    # print "Edge angles: \n", edge_angles

    # Check for angles in 1st quadrant
    for i in range(len(edge_angles)):
        edge_angles[i] = abs(edge_angles[i] % (math.pi / 2))  # want strictly positive answers
    # print "Edge angles in 1st Quadrant: \n", edge_angles

    # Remove duplicate angles
    edge_angles = unique(edge_angles)
    # print "Unique edge angles: \n", edge_angles

    # Test each angle to find bounding box with smallest area
    min_bbox = (0, sys.maxsize, 0, 0, 0, 0, 0, 0)  # rot_angle, area, width, height, min_x, max_x, min_y, max_y  sys.maxint
    print("Testing", len(edge_angles), "possible rotations for bounding box... \n")
    for i in range(len(edge_angles)):

        # Create rotation matrix to shift points to baseline
        # R = [ cos(theta)      , cos(theta-PI/2)
        #       cos(theta+PI/2) , cos(theta)     ]
        R = array([[math.cos(edge_angles[i]), math.cos(edge_angles[i] - (math.pi / 2))], [math.cos(edge_angles[i] + (math.pi / 2)), math.cos(edge_angles[i])]])
        # print "Rotation matrix for ", edge_angles[i], " is \n", R

        # Apply this rotation to convex hull points
        rot_points = dot(R, transpose(hull_points_2d))  # 2x2 * 2xn
        # print "Rotated hull points are \n", rot_points

        # Find min/max x,y points
        min_x = nanmin(rot_points[0], axis=0)
        max_x = nanmax(rot_points[0], axis=0)
        min_y = nanmin(rot_points[1], axis=0)
        max_y = nanmax(rot_points[1], axis=0)
        # print "Min x:", min_x, " Max x: ", max_x, "   Min y:", min_y, " Max y: ", max_y

        # Calculate height/width/area of this bounding rectangle
        width = max_x - min_x
        height = max_y - min_y
        area = width * height
        # print "Potential bounding box ", i, ":  width: ", width, " height: ", height, "  area: ", area

        # Store the smallest rect found first (a simple convex hull might have 2 answers with same area)
        if (area < min_bbox[1]):
            min_bbox = (edge_angles[i], area, width, height, min_x, max_x, min_y, max_y)
        # Bypass, return the last found rect
        # min_bbox = ( edge_angles[i], area, width, height, min_x, max_x, min_y, max_y )

    # Re-create rotation matrix for smallest rect
    angle = min_bbox[0]
    R = array([[math.cos(angle), math.cos(angle - (math.pi / 2))], [math.cos(angle + (math.pi / 2)), math.cos(angle)]])
    # print "Projection matrix: \n", R

    # Project convex hull points onto rotated frame
    proj_points = dot(R, transpose(hull_points_2d))  # 2x2 * 2xn
    # print "Project hull points are \n", proj_points

    # min/max x,y points are against baseline
    min_x = min_bbox[4]
    max_x = min_bbox[5]
    min_y = min_bbox[6]
    max_y = min_bbox[7]
    # print "Min x:", min_x, " Max x: ", max_x, "   Min y:", min_y, " Max y: ", max_y

    # Calculate center point and project onto rotated frame
    center_x = (min_x + max_x) / 2
    center_y = (min_y + max_y) / 2
    center_point = dot([center_x, center_y], R)
    # print "Bounding box center point: \n", center_point

    # Calculate corner points and project onto rotated frame
    corner_points = zeros((4, 2))  # empty 2 column array
    corner_points[0] = dot([max_x, min_y], R)
    corner_points[1] = dot([min_x, min_y], R)
    corner_points[2] = dot([min_x, max_y], R)
    corner_points[3] = dot([max_x, max_y], R)
    # print "Bounding box corner points: \n", corner_points

    # print "Angle of rotation: ", angle, "rad  ", angle * (180/math.pi), "deg"

    return (angle, min_bbox[1], min_bbox[2], min_bbox[3], center_point, corner_points)  # rot_angle, area, width, height, center_point, corner_points


def Get_MBR(points):
    '求平面包围盒'
    hull_points = qhull2D(points)  # Find convex hull 查找凸包
    hull_points = hull_points[::-1]  # Reverse order of points, to match output from other qhull implementations 颠倒点的顺序，以匹配其他qhull实现的输出
    print('凸包点: \n', hull_points, "\n")
    (rot_angle, area, width, height, center_point, corner_points) = minBoundingRect(hull_points)  # Find minimum area bounding rectangle 查找最小面积边界矩形
    print("Minimum area bounding box 最小面积边界框:")
    print("Rotation angle 旋转角:", rot_angle, "rad  (", rot_angle * (180 / math.pi), "deg )")
    print("Width 宽度:", width, " Height 长度:", height, "  Area 面积:", area)
    print("Center point 中心点: \n", center_point)  # numpy array
    print("Corner points 转角点: \n", corner_points, "\n")  # numpy array
    RectangleProperties = np.array([width, height, area])  # 合并矩形属性
    return corner_points, RectangleProperties, center_point


def Get_KB(x1, y1, x2, y2):
    '通过两点式求二维直线的截距和斜率'
    # 计算斜率
    slope = (y2 - y1) / (x2 - x1)
    # 计算截距
    intercept = y1 - slope * x1
    print(f"斜率：{slope}")
    print(f"截距：{intercept}")
    return slope, intercept


def get_cm_all():
    '获得plt和pv支持的colormap列表'
    colormaps = plt.colormaps()  # 列出所有colormap的名字
    j = 0
    for i in colormaps:
        print('colormap', j, ':', i)
        j += 1
    return colormaps


def view_pointclouds(xyz,i,i_name='intensity',colormap='gray',save_path=None):
    '基于pyvista显示点云'
    points = pv.PolyData(xyz)  # 建立pv类
    points[i_name] = i  # 将特征填入类
    # 显示点云
    # 创建一个plotter对象
    plotter = pv.Plotter()
    # 显示点云
    plotter.add_points(points, cmap=colormap, show_scalar_bar=False)
    plotter.background_color = 'white'
    # 添加坐标轴
    plotter.add_axes()
    # 添加坐标轴的刻度
    # plotter.show_grid(color='black', xlabel='X (m)', ylabel='Y (m)', zlabel='Z (m)')  # 显示网格线
    # 显示坐标轴刻度和标签，隐藏网格线
    plotter.show_bounds(
        grid=False,
        location='outer',
        ticks='inside',
        all_edges=False,
        xlabel='X (m)',
        ylabel='Y (m)',
        zlabel='Z (m)'
    )
    plotter.show()
    if save_path:
        # --- 离屏 Plotter 用于截图 ---
        plotter_offscreen = pv.Plotter(off_screen=True)  # 创建离屏 Plotter
        plotter_offscreen.add_points(points, cmap=colormap, show_scalar_bar=False)
        plotter_offscreen.background_color = 'white'
        plotter_offscreen.add_axes()
        # plotter_offscreen.show_grid(color='black', xlabel='X (m)', ylabel='Y (m)', zlabel='Z (m)')
        # 显示坐标轴刻度和标签，隐藏网格线
        plotter_offscreen.show_bounds(
            grid=False,
            location='outer',
            ticks='inside',
            all_edges=False,
            xlabel='X (m)',
            ylabel='Y (m)',
            zlabel='Z (m)'
        )
        arr = plotter_offscreen.screenshot(filename=None, return_img=True)
        img = Image.fromarray(arr, 'RGB')
        dpi = 1200
        img.save(save_path, dpi=(dpi, dpi))
        print(f"图片已保存为: {save_path} (分辨率: {dpi} DPI)")
    # 显示图形



def ReadByPolars(path_file,sep=' '):
    '通过polars读取txt点云'
    xyzi = np.array(pl.read_csv(path_file, has_header=False, separator=sep))  # 读取点云
    return xyzi


def resize_image(input_image_path, output_image_path, new_width, new_height):
    '''
    对图像进行分辨率的修改（https://blog.csdn.net/m0_59799878/article/details/131571082）
    :param input_image_path:读取图像路径
    :param output_image_path:输出图像路径
    :param new_width:输出图像分辨率宽
    :param new_height:输出图像分辨率高
    :return:
    '''
    image = Image.open(input_image_path)
    resized_image = image.resize((new_width, new_height))
    resized_image.save(output_image_path)

def ply2np(input,output=None):
    '读取ply点云并转为numpy格式后存储'
    ply = pf.PlyData.read(input)  # 读取ply
    pcd_ply = ply.elements[0]  # 获取数据属性
    num_names = len(pcd_ply.properties)  # 属性数量
    print('ply点云属性数量',num_names)
    # pcd = []
    for i in range(num_names):  # 对属性数据进行整理
        name_ = pcd_ply.properties[i].name  # 属性名
        print('第',i,'列名称',name_)
        data_ = pcd_ply.data[name_]  # 属性对应数据
        data_ = data_.astype(np.float64)  # 转为numpy
        if i == 0:
            pcd = data_
        else:
            pcd = np.c_[pcd,data_]  # 合并numpy
    print('ply点云数量',len(pcd))
    if output is not None:
        np.savetxt(output,pcd,fmt='%.08f')  # 输出numpy
        print('点云文件已保存在',output)
    return pcd

def fix_plt_CHN():
    '解决matplotlib中文显示乱码问题'
    import matplotlib as mpl
    mpl.rcParams['font.family'] = 'SimHei'  # 设置默认字体为 SimHei
    mpl.rcParams['axes.unicode_minus'] = False  # 解决负号显示问题

def largeCsmall(xyz_l,xyz_s):
    '输出大点云-小点云的点云'
    label_l = xyz_l[:,0]*10+xyz_l[:,1]+xyz_l[:,2]*0.1
    label_s = xyz_s[:,0]*10+xyz_s[:,1]+xyz_s[:,2]*0.1
    id = np.isin(label_l,label_s,invert=True)  # 返回不在的下标
    return xyz_l[id,:]

def AccuracyEvaluation(standard_vehicle,standard_non_vehicle,experiment_vehicle,experiment_non_vehicle):
    '''
    点云精度评价
    :param standard_vehicle: 标准的是点云
    :param standard_non_vehicle: 标准非点云
    :param experiment_vehicle: 测试是点云
    :param experiment_non_vehicle: 测试非点云
    :return: 精度评价结果
    '''
    # 整理前三列数据
    SYV = standard_vehicle[:,2]*100+standard_vehicle[:,1]*10+standard_vehicle[:,0]
    print('标准真数量', len(SYV))
    SNV = standard_non_vehicle[:,2]*100+standard_non_vehicle[:,1]*10+standard_non_vehicle[:,0]
    print('标准假数量', len(SNV))
    EYV = experiment_vehicle[:,2]*100+experiment_vehicle[:,1]*10+experiment_vehicle[:,0]
    print('测试真数量', len(EYV))
    ENV = experiment_non_vehicle[:,2]*100+experiment_non_vehicle[:,1]*10+experiment_non_vehicle[:,0]
    print('测试假数量', len(ENV))
    # 计算True Positives (TP), True Negatives (TN), False Positives (FP), False Negatives (FN)
    # TT = np.isin(EYV, SYV)
    TP = np.sum(np.isin(EYV, SYV))  # 实验提取为车辆且标准为车辆的数量
    print(f"测试为真且标准为真的数量: {TP:.2f}")
    FP = np.sum(np.isin(SNV, EYV))  # 实验提取为车辆但标准为非车辆的数量
    # np.savetxt('FP.txt',standard_non_vehicle[np.isin(SNV, EYV),:],fmt='%.05f')
    print(f"测试为真但标准为假的数量: {FP:.2f}")
    TN = np.sum(np.isin(ENV, SNV))  # 实验提取为非车辆且标准为非车辆的数量
    print(f"测试为假且标准为假的数量: {TN:.2f}")
    FN = np.sum(np.isin(ENV, SYV))  # 实验提取为非车辆但标准为车辆的数量
    print(f"测试为假但标准为真的数量: {FN:.2f}")
    # 计算指标
    type_I_error = FP / (FP + TN)  # 一类误差（Type I Error，假阳性率）
    type_II_error = FN / (FN + TP)  # 二类误差（Type II Error，假阴性率）
    total_error = (FP + FN) / (TP + TN + FP + FN)  # 总误差（Total Error）
    completeness = TP / (TP + FN)  # 完备性（召回率）（Completeness，也称Recall）
    correctness = TP / (TP + FP)  # 正确性（精确率）（Correctness，也称Precision）
    accuracy = (TP + TN) / (TP + TN + FP + FN)  # 准确率（Accuracy）
    quality = TP / (TP + FP + FN)  # 质量（Quality）
    expected_accuracy = ((TP + FN) * (TP + FP) + (TN + FP) * (TN + FN)) / (TP + TN + FP + FN)**2
    kappa = (accuracy - expected_accuracy) / (1 - expected_accuracy)  # kappa
    F1 = 2 * (correctness * completeness) / (correctness + completeness)   # F1指数
    print(f"一类错误 (Type I Error): {type_I_error:.4f}")
    print(f"二类错误 (Type II Error): {type_II_error:.4f}")
    print(f"总错误率 (Total Error): {total_error:.4f}")
    print(f"完备性 (Completeness): {completeness:.4f}")
    print(f"正确性 (Correctness): {correctness:.4f}")
    print(f"准确率 (Accuracy): {accuracy:.4f}")
    print(f"质量 (Quality): {quality:.4f}")
    print(f"卡帕系数 (Kappa coefficient): {kappa:.4f}")   # 存在问题
    print(f"F1指数 (F1 score): {F1:.4f}")
    return np.array([type_I_error,type_II_error,total_error,completeness,correctness,accuracy,quality,kappa,F1])

def read_all_txt(file_path,name=None,sep=' ',limit=5):
    '批量读取文件名带相同字符串的然后合并输出'
    # 获取所有文件名带有 'name' 的 txt 文件
    files = [f for f in os.listdir(file_path) if f.endswith('.txt') and name in f]
    # 使用 Polars 合并所有文件
    j = 0
    for file in files:
        data = ReadByPolars(os.path.join(file_path, file),sep)
        data = data[:,:limit]
        if j==0:
            output = data
        else:
            output = np.r_[output,data]
        j += 1
    return output

def find_first0(arr):
    '找到数组中第一个为0的下标'
    # 找到所有0的位置
    zero_indices = np.where(arr == 0)
    # 第一个0的位置
    first_zero = zero_indices[0][0] if zero_indices[0].size > 0 else -1
    return first_zero

def find_last0(arr):
    '找到数组中最后一个为0的下标'
    # 找到所有0的位置
    zero_indices = np.where(arr == 0)
    # 最后一个0的位置
    last_zero = zero_indices[0][-1] if zero_indices[0].size > 0 else -1
    return last_zero

def ground_ps2DEM(xyz,pixel=1):
    '地面点转DEM,并返回网格坐标'
    x = xyz[:,0]
    y = xyz[:,1]
    z = xyz[:,2]
    # 创建网格
    gird_x,gird_y = np.mgrid[min(x):max(x):pixel,min(y):max(y):pixel]
    gird_z = griddata((x,y),z,(gird_x,gird_y),method='cubic')  # 双三次卷积插值生成DEM
    return gird_z,gird_x,gird_y

def get_buildDSM(xyz,gird_x,gird_y):
    '在建立DEM网格的情况下建立建筑物DSM'
    x = xyz[:,0]
    y = xyz[:,1]
    z = xyz[:,2]
    # 创建DSM网格
    gird_z = griddata((x, y), z, (gird_x, gird_y), method='cubic')  # 双三次卷积插值生成DSM
    return gird_z

def find_continuous_segments(arr):
    """
    找到一维数组中所有连续整数段的起始和终止数
    :param arr: 一维整数数组（已排序）
    :return: 列表，每个元素为 (起始数, 终止数)
    """
    if not arr:
        return []

    segments = []  # 存储所有连续段
    start = arr[0]  # 当前连续段的起始数

    for i in range(1, len(arr)):
        if arr[i] != arr[i - 1] + 1:  # 检测中断点
            segments.append((start, arr[i - 1]))  # 保存当前连续段
            start = arr[i]  # 开始新的连续段

    # 添加最后一个连续段
    segments.append((start, arr[-1]))
    return np.array(segments, dtype=int)

def Read_bin(file,dtype=np.float32,len1=5):
    '读取.bin点云'
    # 读取时需指定数据类型和形状
    loaded_data = np.fromfile(file, dtype=dtype).reshape(-1, len1)  # 假设点云是N×3结构
    return loaded_data



import numpy as np
from numpy.ma.core import indices
from sklearn.neighbors import KDTree
import multiprocessing as mp
from tqdm import tqdm


def getPCA_R_mp(xyz,tree,r_min=0.05,r_max=0.2,dr=0.02,cpu=mp.cpu_count()):
    '求每个点的最小半径'
    num = len(xyz)
    r_all = np.empty(num)

    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    multi_res = pool.starmap_async(getPCA_R, ((i,xyz,tree,r_min,r_max,dr) for i in
                 tqdm(range(num),desc='分配任务计算单个点的适应半径',unit='个点',total=num)))
    j = 0
    for res in tqdm(multi_res.get(),total=num,desc='输出单个点的适应半径'):
        r_all[j] = res[0]
        j += 1
    pool.close()  # 关闭所有进程
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    return r_all

def getPCA_R(i,xyz,tree,r_min,r_max,dr):
    xyz_ = xyz[i, :]  # 当前点
    min_entropy = float('inf')
    optimal_r = r_min
    indices_best = []
    for r_ in np.arange(r_min, r_max, dr):
        indices= tree.query_radius([xyz_], r_)  # 查询半径为0.2内的点
        indices = indices[0]
        indices = indices[indices != i]  # 不包含自己
        xyz_i_r_ = xyz[indices, :]  # 在半径范围内的点
        num_ = len(xyz_i_r_)  # 寻找点的数量
        if num_ >= 2:
            # 构建协方差矩阵并计算特征值
            centered_points = xyz_i_r_ - np.mean(xyz_i_r_, axis=0)
            cov_matrix = np.cov(centered_points, rowvar=False)
            eigenvalues = np.linalg.eigvalsh(cov_matrix)
            eigenvalues = np.sort(eigenvalues)[::-1]  # 降序排序
            # 计算局部维度特征
            L_lambda = (eigenvalues[0] - eigenvalues[1]) / eigenvalues[0]
            P_lambda = (eigenvalues[1] - eigenvalues[2]) / eigenvalues[0]
            S_lambda = eigenvalues[2] / eigenvalues[0]
            # 计算信息熵
            entropy = -L_lambda * np.log(L_lambda) - P_lambda * np.log(P_lambda) - S_lambda * np.log(S_lambda)
            # 最小熵准则
            if entropy < min_entropy:
                min_entropy = entropy
                optimal_r = r_
                indices_best = indices
    return optimal_r,indices

def PCA_new(xyz,r_min=0.05,r_max=0.2,dr=0.02,cpu=mp.cpu_count()):
    '改进PCA算法 ＤＯＩ：１０．１６２５１／ｊ．ｃｎｋｉ．１００９－２３０７．２０２１．１１．０１２．'
    tree = KDTree(xyz)  # 创建树
    # 1.对每个点求最小半径
    r_all = getPCA_R_mp(xyz,tree,r_min,r_max,dr,cpu)
    # 2.计算亲和度
    Affinity = getAffinity_mp(xyz,tree,r_all)
    # 3.计算改进PCA
    return Affinity

def PCA_new_mp(xyz,r_min=0.05,r_max=0.2,dr=0.02,cpu=mp.cpu_count(),k=10):
    '''
    改进PCA算法（全程并行计算版）
    :param xyz: 点云坐标
    :param r_min: 最小半径
    :param r_max: 最大半径
    :param dr: 半径迭代步长
    :param cpu: 并行线程数量
    :param k: K邻域数量
    :return: np.array([L1,L2,L3,VECTOR])*len(xyz)
    '''
    # 准备工作
    num = len(xyz)  # 点云数量
    L123V = np.empty([num,6])  # 前三个为特征值，后三个为最小特征向量
    tree = KDTree(xyz)  # 创建树
    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    Ve = np.empty([num,3])
    Vr = np.empty([num,3])
    # 并行计算
    multi_res = pool.starmap_async(PCA_new_, ((i,xyz,tree,r_min,r_max,dr,k) for i in
                 tqdm(range(num),desc='分配任务计算每个点的PCA',unit='个点',total=num)))
    j = 0
    for res in tqdm(multi_res.get(),total=num,desc='输出单个点的特征值和最小特征向量'):
        Ve[j, :] = res[0]
        Vr[j, :] = res[1]
        j += 1
    pool.close()  # 关闭所有进程
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    return Ve,Vr

def PCA_new_(i,xyz,tree,r_min=0.05,r_max=0.2,dr=0.02,k=10):
    '单点改进PCA算法'
    _,indices = getPCA_R(i, xyz, tree, r_min, r_max, dr)  # 求最适合的半径和下标
    # 计算亲和度和PCA改进矩阵
    xyz_ = xyz[i, :]  # 当前点
    xyz_J = xyz[indices, :]  # 所有J点
    distances = np.linalg.norm(xyz_J - xyz_, axis=1)  # 计算欧式距离
    distance_i_, _ = tree.query([xyz_], k=k + 1)  # 计算I点的第k个点的距离
    num_J = len(xyz_J)
    Ai = np.empty(num_J)  # 一排的亲和度
    centroid = np.mean(xyz_J, axis=0)  # 点云质心
    C = np.zeros([3,3])  # 特征矩阵
    for j in range(num_J):
        xyz_J_ = xyz_J[j, :]
        distance_j_, _ = tree.query([xyz_J_], k=k + 1)  # 计算J点的第k个点的距离
        Aij = np.exp(- (distances[j] ** 2) / (distance_i_[0, -1] * distance_j_[0, -1]))  # 当前两点的亲和度
        Ai[j] = Aij
        # 计算特征值和特征向量
        pj_p = xyz_J_ - centroid  # 临近点减质心
        Cj = Ai[j]*np.outer(pj_p, pj_p)  # 求元素矩阵
        C += Cj
    # 矩阵分解
    eigenvalues, eigenvectors = np.linalg.eig(C)  # 求出矩阵的特征值和特征向量
    sort = eigenvalues.argsort()[::-1]
    eigenvalues = eigenvalues[sort]
    eigenvectors = eigenvectors[:, sort]
    return eigenvalues,eigenvectors[:,-1]  # 返回特征值和最小的特征向量

def getAffinity(xyz,tree,r_all,k=10):
    '单线程计算亲和度'
    num = len(xyz)  # 点的数量
    Affinity = []  # 亲和度容器
    for i in tqdm(range(num),desc='计算两点之间的亲和度',unit='个点',total=num):
        xyz_ = xyz[i,:]  # 当前点
        indices = tree.query_radius([xyz_], r_all[i])  # 需要计算的J点们的下标
        indices = indices[0]
        indices = indices[indices != i]  # 不包含自己
        xyz_J = xyz[indices,:]  # 所有J点
        distances = np.linalg.norm(xyz_J - xyz_, axis=1) # 计算欧式距离
        distance_i_, _ = tree.query([xyz_], k=k + 1)  # 计算I点的第k个点的距离
        num_J = len(xyz_J)
        Ai = np.empty(num_J)  # 一排的亲和度
        for j in range(num_J):
            xyz_J_ = xyz_J[j,:]
            distance_j_, _ = tree.query([xyz_J_], k=k + 1)  # 计算J点的第k个点的距离
            Aij = np.exp(- (distances[j] ** 2) / (distance_i_[0,-1] * distance_j_[0,-1]))  # 当前两点的亲和度
            Ai[j] = Aij
        Affinity.append(Ai)
    return Affinity

def getAffinity_mp(xyz,tree,r_all,k=10,cpu=mp.cpu_count()):
    '并行计算点到点的亲和度'
    # 准备工作
    num = len(xyz)  # 点的数量
    Affinity = []  # 亲和度容器
    pool = mp.Pool(processes=cpu)  # 开启多进程池，数量为cpu
    # 并行计算
    multi_res = pool.starmap_async(getAffinity_, ((i,xyz,tree,r_all,k) for i in
                 tqdm(range(num),desc='分配任务计算每个点的亲和点',unit='个点',total=num)))
    j = 0
    for res in tqdm(multi_res.get(),total=num,desc='输出单个点的适应半径'):
        Affinity.append(res)
        j += 1
    pool.close()  # 关闭所有进程
    pool.join()  # 当所有进程全部计算完毕后，退出并行计算
    return Affinity
def getAffinity_(i,xyz,tree,r_all,k):
    xyz_ = xyz[i, :]  # 当前点
    indices = tree.query_radius([xyz_], r_all[i])  # 需要计算的J点们的下标
    indices = indices[0]
    indices = indices[indices != i]  # 不包含自己
    xyz_J = xyz[indices, :]  # 所有J点
    distances = np.linalg.norm(xyz_J - xyz_, axis=1)  # 计算欧式距离
    distance_i_, _ = tree.query([xyz_], k=k + 1)  # 计算I点的第k个点的距离
    num_J = len(xyz_J)
    Ai = np.empty(num_J)  # 一排的亲和度
    for j in range(num_J):
        xyz_J_ = xyz_J[j, :]
        distance_j_, _ = tree.query([xyz_J_], k=k + 1)  # 计算J点的第k个点的距离
        Aij = np.exp(- (distances[j] ** 2) / (distance_i_[0, -1] * distance_j_[0, -1]))  # 当前两点的亲和度
        Ai[j] = Aij
    return Ai

